"""
File: pset.py
Author: Jeff Martin
Date: 11/1/2021

Copyright © 2021 by Jeffrey Martin. All rights reserved.
Email: jmartin@jeffreymartincomposer.com
Website: https://jeffreymartincomposer.com

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from . import transformations, util
from .pitch import Pitch, PitchClass
import numpy as np
import random

_rng = random.Random()
_rng.seed()


class Sieve:
    """
    Represents a Xenakis sieve. Compatible with all Pitch modulos.
    """
    def __init__(self, tuples, base_pitch: int, pc_mod=12):
        """
        Creates a Sieve
        :param tuples: A collection of tuples to add to the sieve.
        :param base_pitch: Pitch 0 for the sieve. This pitch does not actually have to be
        in the sieve - it will just serve as the 0 reference point.
        :param pc_mod: The pitch-class mod of the Sieve
        """
        self._tuples = set()
        self._intervals = []
        self._pc_mod = pc_mod
        self._period = 0
        self._base_pitch = Pitch(base_pitch, pc_mod)
        self.add_tuples(tuples)

    @property
    def base_pitch(self) -> Pitch:
        """
        The base pitch of the Sieve (pitch 0)
        :return: The base pitch
        """
        return self._base_pitch

    @property
    def intervals(self) -> list:
        """
        The intervallic succession of the Sieve
        :return: The intervallic succession
        """
        return self._intervals

    @property
    def pc_mod(self) -> int:
        """
        The pitch class mod of the Sieve (which specifies the pitch type)
        :return: The pitch class mod
        """
        return self._pc_mod

    @property
    def period(self) -> int:
        """
        The period of the Sieve
        :return: The period
        """
        return self._period

    @property
    def tuples(self) -> set:
        """
        The tuples in the Sieve
        :return: The tuples
        """
        return self._tuples

    def add_tuples(self, *args):
        """
        Adds one or more tuples to the Sieve12
        :param args: One or more tuples
        :return: None
        """
        lcm_list = set()
        if type(args[0]) == set or type(args[0]) == list or type(args[0]) == tuple:
            args = args[0]
        for tup in args:
            self._tuples.add(tup)
        for tup in self._tuples:
            lcm_list.add(tup[0])
        self._period = util.lcm(lcm_list)
        r = self.get_range(Pitch(self._base_p, self._pc_mod), Pitch(self._base_p + self._period, self._pc_mod))
        r = list(r)
        r.sort()
        for i in range(1, len(r)):
            self._intervals.append(r[i].p - r[i - 1].p)

    def get_range(self, p0, p1) -> set:
        """
        Gets all pitches in the sieve between p0 and p1
        :param p0: The low pitch
        :param p1: The high pitch
        :return: A pset
        """
        ps = set()
        p_low = p0.p if type(p0) == Pitch else p0
        p_high = p1.p + 1 if type(p1) == Pitch else p1 + 1
        for j in range(p_low, p_high):
            i = j - self._base_p
            if i >= 0:
                for tup in self._tuples:
                    if i % tup[0] == tup[1]:
                        ps.add(Pitch(j, self._pc_mod))
        return ps

    def intersection(self, sieve) -> 'Sieve':
        """
        Intersects two Sieves
        :param sieve: A Sieve
        :return: A new Sieve. It will have the same base pitch as self.
        """
        t = set()
        for tup1 in self._tuples:
            for tup2 in sieve.tuples:
                if tup1 == tup2:
                    t.add(tup1)
        return Sieve(t, self._base_pitch, self._pc_mod)

    def is_in_sieve(self, p) -> bool:
        """
        Whether or not a pitch or pset is in the sieve
        :param p: A pitch (Pitch or int) or pset
        :return: True or False
        """
        ps = None
        if type(p) == set:
            ps = p
        elif type(p) == Pitch:
            ps = {p}
        elif type(p) == int:
            ps = {Pitch(p, self._pc_mod)}
        for q in ps:
            i = q.p - self._base_p
            if i < 0:
                return False
            else:
                for tup in self._tuples:
                    if i % tup[0] == tup[1]:
                        break
                else:
                    return False
        return True

    def union(self, sieve) -> 'Sieve':
        """
        Unions two Sieves
        :param sieve: A Sieve
        :return: A new Sieve. It will have the same base pitch as self.
        """
        t = set()
        for tup in self._tuples:
            t.add(tup)
        for tup in sieve.tuples:
            t.add(tup)
        return Sieve(t, self._base_pitch, self._pc_mod)


def calculate_pm_similarity(pset1: set, pset2: set, ic_roster1=None, ic_roster2=None) -> tuple:
    """
    Gets the pitch-measure (PM) similarity between pset1 and pset2
    :param pset1: A pset
    :param pset2: A pset
    :param ic_roster1: The ic_roster for pset 1. If None, will be calculated.
    :param ic_roster2: The ic_roster for pset 2. If None, will be calculated.
    :return: The PM similarity as a tuple of integers
    *Compatible with all Pitch modulos
    """
    cint = len(pset1.intersection(pset2))
    ic_shared = 0
    if ic_roster1 is None:
        ic_roster1 = get_ic_roster(pset1)
    if ic_roster2 is None:
        ic_roster2 = get_ic_roster(pset2)
    for ic in ic_roster1:
        if ic in ic_roster2:
            if ic_roster1[ic] < ic_roster2[ic]:
                ic_shared += ic_roster1[ic]
            else:
                ic_shared += ic_roster2[ic]
    return (cint, ic_shared)


def generate_random_pset_realizations(pcset: set, lower_boundary: int, upper_boundary: int, num_realizations: int=1, num_duplicate_pitches: int=0, filter_func=None):
    """
    Generates random pset realizations of a given pcset, 
    within the specified upper and lower boundaries
    :param pcset: The pcset to realize
    :param lower_boundary: The lower boundary
    :param upper_boundary: The upper boundary
    :param num_realizations: The number of random realizations to generate
    :param num_duplicate_pitches: The number of additional duplicate pitches to include (for doubling)
    :param filter_func: A function for filtering the pset realizations to force them to match specified criteria
    :return: One or more random pset realizations of the pcset within the given boundaries. If the number of realizations
    is greater than 1, returns a list of psets. Otherwise returns a single pset.
    *Compatible with all Pitch modulos
    """
    if len(pcset) == 0:
        return set()
    else:
        mod = next(iter(pcset)).mod
        lowest_boundary_note = lower_boundary % mod
        
        # Generate all of the possible pitch realizations for each pitch-class
        # within the range provided
        pitch_choices = []
        for pc in pcset:
            choices = []
            candidate_pitch = (pc.pc - lowest_boundary_note) % mod + lower_boundary           
            in_range = True
            while in_range:
                if candidate_pitch > upper_boundary:
                    in_range = False
                else:
                    choices.append(Pitch(candidate_pitch, mod))
                candidate_pitch += mod
            pitch_choices.append(choices)

        # Generate pset realizations
        realizations = {}
        i = 0
        while len(realizations) < num_realizations and i < 10 * num_realizations:
            # Build the current realization
            realization = set()
            unused_pitches = []
            for bucket in pitch_choices:
                if len(bucket) > 0:
                    idx = _rng.randrange(0, len(bucket))
                    realization.add(bucket[idx])
                    unused_pitches += bucket[:idx] + bucket[idx+1:]

            # Add duplicate pitches 
            for i in range(num_duplicate_pitches):
                idx = _rng.randrange(0, len(unused_pitches))
                realization.add(unused_pitches[idx])
                del unused_pitches[idx]

            if filter_func is None:
                realizations[str(realization)] = realization
            elif filter_func(realization):
                realizations[str(realization)] = realization
            i += 1
        realizations = [realizations[key] for key in realizations]

        if len(realizations) == 1:
            return realizations[0]
        else:
            return realizations


def get_fb_class(pset: set, p0: int) -> list:
    """
    Gets the FB-class of a pset
    :param pset: The pset
    :param p0: The lowest pitch
    :return: The FB-class as a list of integers
    *Compatible with all Pitch modulos
    """
    intlist = []
    if len(pset) > 0 and p0 >= 0:
        mod = pset[0].mod
        for p in pset:
            intlist.append((p.p - p0) % mod)
        intlist.sort()
        if len(intlist) > 0:
            del intlist[0]
    return intlist


def get_ic_matrix(pset: set) -> np.ndarray:
    """
    Gets the pitch ic-matrix
    :param pset: The pset
    :return: The ic-matrix as a list of lists
    *Compatible with all Pitch modulos
    """
    mx = np.empty((len(pset), len(pset)))
    pseg = list(pset)
    pseg.sort()
    for i in range(mx.shape[0]):
        for j in range(mx.shape[1]):
            mx[i][j] = abs(pseg[i].p - pseg[j].p)
    return mx


def get_ic_roster(pset: set) -> dict:
    """
    Gets the pitch ic-roster
    :param pset: The pset
    :return: The ic-roster as a dictionary
    *Compatible with all Pitch modulos
    """
    pseg = list(pset)
    roster = {}
    pseg.sort()
    for i in range(len(pseg) - 1, -1, -1):
        for j in range(i - 1, -1, -1):
            interval = abs(pseg[i].p - pseg[j].p)
            if interval not in roster:
                roster[interval] = 1
            else:
                roster[interval] += 1
    return roster


def get_pcint_class(pset: set) -> list:
    """
    Gets the PCINT-class of a pset
    :param pset: The pset
    :return: The PCINT-class as a list of integers
    *Compatible with all Pitch modulos
    """
    intlist = []
    if len(pset) > 0:
        pseg = list(pset)
        pseg.sort()
        mod = pseg[0].mod
        for i in range(1, len(pseg)):
            intlist.append((pseg[i].p - pseg[i - 1].p) % mod)
    return intlist


def get_set_class(pset: set) -> list:
    """
    Gets the set-class of a pset
    :param pset: The pset
    :return: The set-class as a list of integers
    *Compatible with all Pitch modulos
    """
    pseg = list(pset)
    pseg.sort()
    intlist = []
    for i in range(1, len(pseg)):
        intlist.append(pseg[i].p - pseg[i - 1].p)
    return intlist


def invert(pset: set) -> set:
    """
    Inverts a pset
    :param pset: The pset
    :return: The inverted pset
    *Compatible with all Pitch modulos
    """
    pset2 = set()
    if len(pset) > 0:
        mod = next(iter(pset)).mod
        for p in pset:
            pset2.add(Pitch(p.p * -1, mod))
    return pset2


def make_pset12(*args) -> set:
    """
    Makes a pset
    :param *args: Pitches
    :return: A pset
    *Compatible only with chromatic psegs
    """
    if type(args[0]) == list:
        args = args[0]
    return {Pitch(p, 12) for p in args}


def make_pset24(*args) -> set:
    """
    Makes a pset
    :param *args: Pitches
    :return: A pset
    *Compatible only with microtonal psegs
    """
    if type(args[0]) == list:
        args = args[0]
    return {Pitch(p, 24) for p in args}


def subsets(pset: set) -> list:
    """
    Gets all subsets of a pset, using the bit masking solution from
    https://afteracademy.com/blog/print-all-subsets-of-a-given-set
    :param pset: A pset
    :return: A list containing all subsets of the pset
    *Compatible with all Pitch modulos
    """
    total = 2 ** len(pset)
    t = type(next(iter(pset)))
    sub = []
    pseg = list(pset)
    pseg.sort()
    for index in range(total):
        sub.append([])
        for i in range(len(pset)):
            if index & (1 << i):
                sub[index].append(t(pseg[i].p))
    sub.sort()
    return sub


def to_pcset(pset: set) -> set:
    """
    Makes a pcset out of a pset
    :param pset: A pset
    :return: A pcset
    *Compatible with all Pitch modulos
    """
    if len(pset) > 0:
        mod = next(iter(pset)).mod
        return {PitchClass(p.pc, mod) for p in pset}
    else:
        return set()


def transform(pset: set, transformation: transformations.UTO) -> set:
    """
    Transforms a pset
    :param pset: A pset
    :param transformation: A transformation
    :return: The transformed set
    *Compatible with all Pitch modulos
    """
    pset2 = set()
    if len(pset) > 0:
        mod = next(iter(pset)).mod
        for p in pset:
            pset2.add(Pitch(p.p * transformation[1] + transformation[0], mod))
    return pset2


def transpose(pset: set, n: int) -> set:
    """
    Transposes a pset
    :param pset: The pset
    :param n: The index of transposition
    :return: The transposed pset
    *Compatible with all Pitch modulos
    """
    pset2 = set()
    if len(pset) > 0:
        mod = next(iter(pset)).mod
        for p in pset:
            pset2.add(Pitch(p.p + n, mod))
    return pset2



"""
File: group.py
Author: Jeff Martin
Date: 12/23/2021

Copyright © 2021 by Jeffrey Martin. All rights reserved.
Email: jmartin@jeffreymartincomposer.com
Website: https://jeffreymartincomposer.com

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from queue import Queue
from . import transformations
from .transformations import UTO
from .pitch import PitchClass


class OperatorGroup:
    """
    Represents a group of operators. Only compatible with mod 12 and mod 24 systems.
    """
    def __init__(self, utos: list = None, mod: int = 12):
        """
        Creates an OperatorGroup.
        :param utos: UTOs
        :param mod: The number of pcs in the system of the group (chromatic: 12, microtonal: 24)
        """
        self._MNUM_12 = {1, 5, 7, 11}
        self._MNUM_24 = {1, 5, 7, 11, 13, 17, 19, 23}
        self._name = ""
        self._num_pcs = mod
        if mod == 12:
            self._operators = [[] for i in range(len(self._MNUM_12))]
        elif mod == 24:
            self._operators = [[] for i in range(len(self._MNUM_24))]
        self._utos = set()
        if utos is not None:
            self.load_utos(utos)

    def __contains__(self, uto: UTO):
        return uto in self._utos

    def __iter__(self):
        return (uto for uto in self._utos)

    def __list__(self):
        uto_list = list(self._utos)
        return uto_list
        
    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name

    @property
    def name(self) -> str:
        """
        Gets the group name.
        :return: The group name
        """
        group_name = "G"
        if self._num_pcs == 12:
            for i in range(len(self._MNUM_12) - 1):
                if len(self._operators[i]) == 12:
                    group_name += "*"
                else:
                    for uto in self._operators[i]:
                        group_name += str(uto[0])
                group_name += "/"
            if len(self._operators[len(self._MNUM_12) - 1]) == 12:
                group_name += "*"
            else:
                for uto in self._operators[len(self._MNUM_12) - 1]:
                    group_name += str(uto[0])
        else:
            for i in range(len(self._MNUM_24) - 1):
                if len(self._operators[i]) == 24:
                    group_name += "*"
                else:
                    for j in range(len(self._operators[i]) - 1):
                        group_name += f"{self._operators[i][j][0]},"
                    if len(self._operators[i]) > 0:
                        group_name += str(self._operators[i][len(self._operators[i]) - 1][0])
                group_name += "/"
            if len(self._operators[len(self._MNUM_24) - 1]) == 24:
                group_name += "*"
            else:
                for j in range(len(self._operators[len(self._MNUM_24) - 1]) - 1):
                    group_name += f"{self._operators[len(self._MNUM_24) - 1][j][0]},"
                if len(self._operators[len(self._MNUM_24) - 1]) > 0:
                    group_name += str(self._operators[len(self._MNUM_24) - 1][len(self._operators[len(self._MNUM_24) - 1]) - 1][0])
        return group_name

    @property
    def utos(self) -> set:
        """
        The set of UTOs in the group
        :return: The set of UTOs
        """
        return self._utos

    def get_orbits(self) -> list:
        """
        Gets the orbits of the group.
        :return: The orbits, as a list of sets
        """
        orbits = []
        u = None
        if self._num_pcs == 12:
            u = {PitchClass(i, 12) for i in range(self._num_pcs)}
        elif self._num_pcs == 24:
            u = {PitchClass(i, 24) for i in range(self._num_pcs)}
        while len(u) > 0:
            orbit = set()
            q = Queue()
            pc = next(iter(u))
            q.put(pc)
            orbit.add(pc)
            u.remove(pc)
            while not q.empty():
                pc = q.get()
                for op in self._utos:
                    tr = op.transform(pc)
                    if tr not in orbit:
                        orbit.add(tr)
                        q.put(tr)
                        u.remove(tr)
            orbits.append(orbit)

        return orbits

    def left_coset(self, uto) -> list:
        """
        Gets a left coset of the group.
        :param uto: A UTO
        :return: The left coset
        """
        coset = []
        for u in self._utos:
            coset.append(transformations.left_multiply_utos(uto, u))
        coset.sort()
        return coset

    def load_utos(self, utos: list):
        """
        Loads UTOs into the group.
        :param utos: UTOs
        """
        self._utos = set()
        for uto in utos:
            match uto[1]:
                case 1:
                    self._operators[0].append(uto)
                case 5:
                    self._operators[1].append(uto)
                case 7:
                    self._operators[2].append(uto)
                case 11:
                    self._operators[3].append(uto)
                case 13:
                    self._operators[4].append(uto)
                case 17:
                    self._operators[5].append(uto)
                case 19:
                    self._operators[6].append(uto)
                case 23:
                    self._operators[7].append(uto)
            self._utos.add(uto)
        for li in self._operators:
            li.sort()

    def right_coset(self, uto: UTO) -> list:
        """
        Gets a right coset of the group.
        :param uto: A UTO
        :return: The right coset
        """
        coset = []
        for u in self._utos:
            coset.append(transformations.left_multiply_utos(u, uto))
        coset.sort()
        return coset



"""
File: util.py
Author: Jeff Martin
Date: 11/13/2021

Copyright © 2021 by Jeffrey Martin. All rights reserved.
Email: jmartin@jeffreymartincomposer.com
Website: https://jeffreymartincomposer.com

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""


def factor(n) -> list:
    """
    Factors a positive integer
    :param n: An integer
    :returns: A list of factors, in sorted order, including duplicates
    """
    factors = []
    d = 1
    while d <= int(n ** 0.5):
        if n % d == 0:
            factors.append(d)
            n //= d
        else:
            d += 1
        if d == 1:
            d += 1
        # if d > int(n ** 0.5):
        #     factors.append(n)
    factors.append(n)
    # factors.sort()
    return factors


def lcm(integers) -> int:
    """
    Computes the LCM of a list of positive integers
    :param integers: A list of positive integers
    :return: The LCM
    """
    factors = {}  # A dictionary of individual factors and their multiplicities
    multiple = 1  # The LCM

    for num in integers:
        cur_factors = factor(num)  # The factors of the current number
        current = 1  # The current factor we are considering
        count = 0  # The number of occurrences of that factor
        for i in range(len(cur_factors)):
            # If we found another occurrence of that factor, increase the count
            if cur_factors[i] == current:
                count += 1
            # Otherwise record the count and move on
            else:
                if current not in factors:
                    factors[current] = count
                elif factors[current] < count:
                    factors[current] = count
                current = cur_factors[i]
                count = 1
            # If we are done, record the count of the last factor
            if i + 1 == len(cur_factors):
                if current not in factors:
                    factors[current] = count
                elif factors[current] < count:
                    factors[current] = count

    # Compute the LCM
    for item in factors:
        multiple *= item ** factors[item]
    # print(multiple)
    return multiple


def map_to_chromatic(scale_map, sequence) -> list:
    """
    Maps one sequence of items to another sequence of items. This is useful for doing 
    things like mapping scale degrees 0-6 to actual chromatic pitches. You only need
    to provide a map for one octave, and all transpositions will be accounted for.
    :param scale_map: The scale map
    :param sequence: The sequence to map
    :return: The mapped sequence
    """
    CMOD = 12
    smod = len(scale_map)
    sequence2 = []
    for p in sequence:
        pc = ((p % smod) + smod) % smod
        p2 = (p // smod) * CMOD + scale_map[pc]
        sequence2.append(p2)
    return sequence2


def norgard(n: int) -> list:
    """
    Generates the first n numbers of OEIS A004718 (Per Nørgård's infinity series)
    :param n: The number of terms to compute
    :return: The series
    """
    n_list = [0 for i in range(n)]
    i = 0
    m = 0
    while m < n:
        m = 2 * i
        if m < n:
            n_list[m] = -n_list[i]
        m += 1
        if m < n:
            n_list[m] = n_list[i] + 1
        i += 1
    return n_list



"""
File: pcseg.py
Author: Jeff Martin
Date: 10/30/2021

Copyright © 2021 by Jeffrey Martin. All rights reserved.
Email: jmartin@jeffreymartincomposer.com
Website: https://jeffreymartincomposer.com

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from . import pcset, tables, transformations
from .pitch import PitchClass
import random

_rng = random.Random()
_rng.seed()


def adjacent_search(pcseg: list, corpus: list) -> list:
    """
    Performs an adjacent search on a corpus of pcsegs. Returns all pcsegs in the corpus that contain pcseg as an ordered sublist.
    :param pcseg: A pcseg to search for
    :param corpus: The corpus of pcsegs to search
    :return: A sublist of pcsegs from the corpus that have pcseg in them. If an entry in the original corpus contains pcseg more than once,
    it will only be included once in this sublist.
    If no entries have pcseg in them, returns an empty list.
    """
    new_corpus = []
    for entry in corpus:
        for i in range(0, len(entry) - len(pcseg)):
            found = True
            for j in range(0, len(pcseg)):
                if pcseg[j] != entry[i+j]:
                    found = False
                    break
            if found:
                new_corpus.append(entry)
                break
    return new_corpus


def are_combinatorial2(row1: list, row2: list) -> bool:
    """
    Determines if two rows are hexachordally combinatorial.
    :param row1: A row
    :param row2: A row
    :return: True or False
    *Compatible only with chromatic pcsegs
    """
    if len(row1) != 12 or len(row2) != 12:
        return False
    s1 = set(row1[:6] + row2[:6])
    s2 = set(row1[6:] + row2[6:])
    if len(s1) == len(s2) == 12:
        return True
    else:
        return False


def are_combinatorial3(row1: list, row2: list, row3: list) -> bool:
    """
    Determines if three rows are tetrachordally combinatorial.
    :param row1: A row
    :param row2: A row
    :param row3: A row
    :return: True or False
    *Compatible only with chromatic pcsegs
    """
    if len(row1) != 12 or len(row2) != 12:
        return False
    s1 = set(row1[:4] + row2[:4] + row3[:4])
    s2 = set(row1[4:8] + row2[4:8] + row3[4:8])
    s3 = set(row1[8:] + row2[8:] + row3[8:])
    if len(s1) == len(s2) == len(s3) == 12:
        return True
    else:
        return False


def are_combinatorial4(row1: list, row2: list, row3: list, row4: list) -> bool:
    """
    Determines if four rows are trichordally combinatorial.
    :param row1: A row
    :param row2: A row
    :param row3: A row
    :param row4: A row
    :return: True or False
    *Compatible only with chromatic pcsegs
    """
    if len(row1) != 12 or len(row2) != 12:
        return False
    s1 = set(row1[:3] + row2[:3] + row3[:3] + row4[:3])
    s2 = set(row1[3:6] + row2[3:6] + row3[3:6] + row4[3:6])
    s3 = set(row1[6:9] + row2[6:9] + row3[6:9] + row4[6:9])
    s4 = set(row1[9:] + row2[9:] + row3[9:] + row4[9:])
    if len(s1) == len(s2) == len(s3) == len(s4) == 12:
        return True
    else:
        return False


def bip_n(imb: list) -> list:
    """
    Gets the BIP_n of a list of set-classes generated by IMB_n.
    :param imb: The IMB_n list
    :return: The BIP_n
    *Compatible only with chromatic pcsegs
    """
    bip = []
    if len(imb) > 0:
        for sc in imb:
            name = sc.name_forte.split('-')[1]
            name = name.replace("Z", "")
            bip.append(int(name))
        bip.sort()
    return bip


def create_ormap(pcseg: list) -> dict:
    """
    Creates the ORMAP of a row.
    :param pcseg: A row
    :return: The ORMAP
    *Compatible with all PitchClass modulos    
    """
    omap = {}
    mod = pcseg[0].mod
    for i in range(len(pcseg)):
        omap[PitchClass(pcseg[i].pc, mod)] = i
    return omap


def generate_pcseg12_from_interval_list(interval_list: list, starting_pc=None) -> list:
    """
    Generates a pcseg from an interval list.
    :param interval_list: The interval list
    :param starting_pc: The starting pitch-class. If None, a random starting pitch-class is used.
    :return: A pcseg
    *Compatible only with chromatic pcsegs
    """
    if starting_pc is None:
        pcseg = [PitchClass(_rng.randrange(12), 12)]
        for i in range(len(interval_list)):
            pcseg.append(PitchClass(pcseg[i].pc + interval_list[i], 12))
        return pcseg
    else:
        pcseg = [PitchClass(starting_pc, 12)]
        for i in range(len(interval_list)):
            pcseg.append(PitchClass(pcseg[i].pc + interval_list[i], 12))
        return pcseg


def generate_pcseg24_from_interval_list(interval_list: list, starting_pc=None) -> list:
    """
    Generates a pcseg from an interval list.
    :param interval_list: The interval list
    :param starting_pc: The starting pitch-class. If None, a random starting pitch-class is used.
    :return: A pcseg
    *Compatible only with microtonal pcsegs
    """
    if starting_pc is None:
        pcseg = [PitchClass(_rng.randrange(24), 24)]
        for i in range(len(interval_list)):
            pcseg.append(PitchClass(pcseg[i].pc + interval_list[i], 24))
        return pcseg
    else:
        pcseg = [PitchClass(starting_pc, 24)]
        for i in range(len(interval_list)):
            pcseg.append(PitchClass(pcseg[i].pc + interval_list[i], 24))
        return pcseg


def generate_random_all_interval_row(starting_pc=None) -> list:
    """
    Generates a random all-interval row.
    :param starting_pc: The starting pitch-class. If None, a random starting pitch-class is used.
    :return: An all-interval row
    *Compatible only with chromatic pcsegs
    """
    name_tables = tables.create_tables_eleven_interval()
    row = []

    # Establish the starting pitch of the row
    if starting_pc is not None:
        if type(starting_pc) == PitchClass:
            row.append(PitchClass(starting_pc.pc, 12))
        else:
            row.append(PitchClass(starting_pc, 12))
    else:
        row.append(PitchClass(_rng.randrange(12), 12))
    generator = name_tables["elevenIntervalRowGenerators"][_rng.randrange(
        len(name_tables["elevenIntervalRowGenerators"]))]

    # Randomly choose to invert the row generator
    if _rng.randrange(1) == 1:
        generator = [i * 11 % 12 for i in generator]

    # Build the row
    for i in range(1, 12):
        row.append(PitchClass(row[i - 1].pc + generator[i - 1], 12))
    return row


def generate_random_all_trichord_row(starting_pc=None) -> list:
    """
    Generates a random all-trichord row.
    :param starting_pc: The starting pitch-class. If None, a random starting pitch-class is used.
    :return: An all-trichord row
    *Compatible only with chromatic pcsegs
    """
    name_tables = tables.create_tables_all_trichord()
    if starting_pc is not None:
        if type(starting_pc) == PitchClass:
            starting_pc = starting_pc.pc
    else:
        starting_pc = 0
    row = [PitchClass(pc + starting_pc, 12) for pc in name_tables["allTrichordRows"][_rng.randrange(
        len(name_tables["allTrichordRows"]))]]

    # Randomly choose to invert the row generator
    if _rng.randrange(1) == 1:
        row = invert(row)

    return row


def generate_random_all_trichord_babbitt_row(starting_pc=None) -> list:
    """
    Generates a random all-trichord (Babbitt) row.
    :param starting_pc: The starting pitch-class. If None, a random starting pitch-class is used.
    :return: An all-trichord (Babbitt) row
    *Compatible only with chromatic pcsegs
    """
    name_tables = tables.create_tables_all_trichord_babbitt()
    if starting_pc is not None:
        if type(starting_pc) == PitchClass:
            starting_pc = starting_pc.pc
    else:
        starting_pc = 0
    row = [PitchClass(pc + starting_pc, 12) for pc in name_tables["allTrichordBabbittRows"][_rng.randrange(
        len(name_tables["allTrichordBabbittRows"]))]]

    # Randomly choose to invert the row generator
    if _rng.randrange(1) == 1:
        row = invert(row)

    return row


def generate_random_pcseg12(length: int, non_duplicative: bool = False, starting_pc=None) -> list:
    """
    Generates a random pcseg.
    :param length: The length of the pcseg
    :param non_duplicative: Whether or not duplicate pcs may occur (must be True to generate a row)
    :param starting_pc: The starting pitch-class. If None, a random starting pitch-class is used.
    :return: A random pcseg
    *Compatible only with chromatic pcsegs
    """
    pcseg = [PitchClass(_rng.randrange(12) if starting_pc is None else starting_pc, 12)]
    if non_duplicative and 0 < length <= 12:
        pcs = [i for i in range(12)]
        del pcs[pcseg[0].pc]
        for i in range(length - 1):
            j = _rng.randrange(len(pcs))
            pcseg.append(PitchClass(pcs[j], 12))
            del pcs[j]
    elif not non_duplicative and 0 < length:
        for i in range(length - 1):
            pcseg.append(PitchClass(_rng.randrange(12), 12))
    else:
        raise ValueError("Invalid length")
    return pcseg


def generate_random_pcseg24(length: int, non_duplicative: bool = False, starting_pc=None) -> list:
    """
    Generates a random pcseg.
    :param length: The length of the pcseg
    :param non_duplicative: Whether or not duplicate pcs may occur (must be True to generate a row)
    :param starting_pc: The starting pitch-class. If None, a random starting pitch-class is used.
    :return: A random pcseg
    *Compatible only with microtonal pcsegs
    """
    pcseg = [PitchClass(_rng.randrange(24) if starting_pc is None else starting_pc, 24)]
    if non_duplicative and 0 < length <= 24:
        pcs = [i for i in range(24)]
        del pcs[pcseg[0].pc]
        for i in range(length - 1):
            j = _rng.randrange(len(pcs))
            pcseg.append(PitchClass(pcs[j], 24))
            del pcs[j]
    elif not non_duplicative and 0 < length:
        for i in range(length - 1):
            pcseg.append(PitchClass(_rng.randrange(24), 24))
    else:
        raise ValueError("Invalid length")
    return pcseg


def generate_random_pcseg_from_pcset(pcset: set) -> list:
    """
    Generates a random pcseg from a pcset.
    :param pcset: A pcset
    :return: A pcseg
    *Compatible with all PitchClass modulos
    """
    pcseg = []
    if len(pcset) > 0:
        setseg = list(pcset)
        mod = setseg[0].mod
        for i in range(len(setseg)):
            j = _rng.randrange(len(setseg))
            pcseg.append(PitchClass(setseg[j].pc, mod))
            del setseg[j]
    return pcseg


def generate_random_ten_trichord_row(starting_pc=None) -> list:
    """
    Generates a random ten-trichord row.
    :param starting_pc: The starting pitch-class. If None, a random starting pitch-class is used.
    :return: A ten-trichord row
    *Compatible only with chromatic pcsegs
    """
    name_tables = tables.create_tables_ten_trichord()
    if starting_pc is not None:
        if type(starting_pc) == PitchClass:
            starting_pc = starting_pc.pc
    else:
        starting_pc = 0
    row = [PitchClass(pc + starting_pc, 12) for pc in name_tables["tenTrichordRows"][_rng.randrange(
        len(name_tables["tenTrichordRows"]))]]

    # Randomly choose to invert the row generator
    if _rng.randrange(1) == 1:
        row = invert(row)

    return row


def get_intervals(pcseg: list) -> list:
    """
    Gets the interval sequence of a pcseg.
    :param pcseg: The pcseg
    :return: The interval sequence
    *Compatible with all PitchClass modulos    
    """
    intervals = []
    for i in range(1, len(pcseg)):
        intervals.append(pcseg[i].pc - pcseg[i - 1].pc)
    return intervals


def get_row_class(row: list) -> list:
    """
    Gets all of the rows in a row-class.
    :param row: A row
    :return: The row-class, as a list of rows
    *Compatible with PitchClasses mod 12 and 24
    """
    row_class1 = []
    if len(row) == 12 and row[0].mod == 12:
        ros = transformations.get_otos12()
        mod = 12
    elif len(row) == 24 and row[0].mod == 24:
        ros = transformations.get_otos24()
        mod = 24
    else:
        return row_class1

    row_class = {}
    for t in ["", "I", "R", "RI"]:
        for i in range(mod):
            transform = ros[f"T{i}{t}"].transform(row)
            row_name = str(transform)
            if row_name not in row_class:
                row_class[row_name] = transform
    for row_name in row_class:
        row_class1.append(row_class[row_name])
    
    return row_class1


def get_row_dsym(row: list) -> int:
    """
    Gets the degree of symmetry of a row.
    :param row: A row
    :return: The degree of symmetry of the row
    *Compatible with PitchClasses mod 12 and 24
    """
    if len(row) == 12 and row[0].mod == 12:
        return 48 // len(get_row_class(row))
    elif len(row) == 24 and row[0].mod == 24:
        return 96 // len(get_row_class(row))
    else:
        return 0


def imb_n(pcseg: list, n: int) -> list:
    """
    Gets the IMB_n of a pcseg. The IMB_n is the segment of imbricated set-classes of cardinality n.
    :param pcseg: The pcseg
    :param n: The cardinality of imbrication
    :return: The IMB_n
    *Compatible with all PitchClass modulos
    """
    imb = set()
    scs = []
    if len(pcseg) > 0 and n > 0:
        mod = pcseg[0].mod
        for i in range(len(pcseg) + 1 - n):
            for j in range(i, i + n):
                imb.add(pcseg[j])
            scs.append(pcset.SetClass(imb, mod))
            imb.clear()
    return scs


def invert(pcseg: list) -> list:
    """
    Inverts a pcseg.
    :param pcseg: The pcseg
    :return: The inverted pcseg
    *Compatible with all PitchClass modulos
    """
    pcseg2 = []
    if len(pcseg) > 0:
        mod = pcseg[0].mod
        for pc in pcseg:
            pcseg2.append(PitchClass(pc.pc * -1, mod))
    return pcseg2


def is_all_interval_row(pcseg: list) -> bool:
    """
    Determines if a pcseg is an all-interval row.
    :param pcseg: The pcseg
    :return: Whether or not the pcseg is an all-interval row
    *Compatible with PitchClasses mod 12 and 24
    """
    pcs = {pc.pc for pc in pcseg}
    if len(pcseg) == 12 and pcseg[0].mod == 12:
        if len(pcs) != len(pcseg):
            return False
        else:
            ints = {(pcseg[i].pc - pcseg[i-1].pc) % 12 for i in range(1, len(pcseg))}
            if len(ints) == 11:
                return True
    elif len(pcseg) == 24 and pcseg[0].mod == 24:
        if len(pcs) != len(pcseg):
            return False
        else:
            ints = {(pcseg[i].pc - pcseg[i-1].pc) % 24 for i in range(1, len(pcseg))}
            if len(ints) == 23:
                return True
    return False


def is_all_trichord_row(pcseg: list) -> bool:
    """
    Determines if a pcseg is an all-trichord row.
    :param pcseg: The pcseg
    :return: Whether or not the pcseg is an all-trichord row
    *Compatible only with chromatic pcsegs
    """
    if not is_row(pcseg):
        return False
    row1 = [pcseg[i % 12] for i in range(14)]
    imb3 = imb_n(row1, 3)
    if len(set(imb3)) == 12:
        return True
    else:
        return False


def is_all_trichord_babbitt_row(pcseg: list) -> bool:
    """
    Determines if a pcseg is an "all-trichord" row (after Babbitt). This kind of row
    contains ten unique imbricated trichords, none of which are set-classes [036] or [048].
    Rotation is not considered.
    :param pcseg: The pcseg
    :return: Whether or not the pcseg is an "all-trichord" row (after Babbitt)
    *Compatible only with chromatic pcsegs
    """
    if not is_row(pcseg):
        return False
    imb3 = imb_n(pcseg, 3)
    if len(set(imb3)) == 10:
        for sc in imb3:
            if sc.name_prime == "[036]" or sc.name_prime == "[048]":
                return False
        return True
    else:
        return False


def is_link_chord(pcseg: list) -> bool:
    """
    Determines if a row is a Link chord (an all-interval row containing the all-trichord hexachord as a subset).
    :param pcseg: A pcset
    :return: True or False
    *Compatible only with chromatic pcsegs
    """
    if is_all_interval_row(pcseg):
        imb = imb_n(pcseg, 6)
        for sc in imb:
            if sc.name_forte == "6-Z17":
                return True
    return False


def is_row(pcseg: list) -> bool:
    """
    Determines if a pcseg is a row.
    :param pcseg: The pcseg
    :return: Whether or not the pcseg is a row
    *Compatible with PitchClasses mod 12 and 24
    """
    pcs = set([pc.pc for pc in pcseg])
    if (len(pcseg) == 12 and pcseg[0].mod == 12) or (len(pcseg) == 24 and pcseg[0].mod == 24):
        if len(pcs) != len(pcseg):
            return False
        else:
            return True
    else:
        return False


def is_row_generator(rgen: list) -> bool:
    """
    Determines if a row generator is valid.
    :param rgen: A row generator
    :return: True if the row generator is valid; false otherwise
    *Compatible only with chromatic pcsegs
    """
    for i in range(len(list) - 1):
        rgen_sum = rgen[i]
        for j in range(i + 1, len(list)):
            rgen_sum += rgen[j]
            if rgen_sum % 12 == 0:
                return False
    return True


def is_ten_trichord_row(pcseg: list) -> bool:
    """
    Determines if a pcseg is a ten-trichord row. This kind of row
    contains ten unique imbricated trichords. Rotation is not considered.
    :param pcseg: The pcseg
    :return: Whether or not the pcseg is an "all-trichord" row (after Babbitt)
    *Compatible only with chromatic pcsegs
    """
    if not is_row(pcseg):
        return False
    imb3 = imb_n(pcseg, 3)
    if len(set(imb3)) == 10:
        return True
    else:
        return False


def make_pcseg12(*args) -> list:
    """
    Makes a pcseg.
    :param args: Pcs
    :return: A pcseg
    *Compatible only with chromatic pcsegs
    """
    if type(args[0]) == list:
        args = args[0]
    return [PitchClass(pc, 12) for pc in args]


def make_pcseg24(*args) -> list:
    """
    Makes a pcseg.
    :param args: Pcs
    :return: A pcseg
    *Compatible only with microtonal pcsegs
    """
    if type(args[0]) == list:
        args = args[0]
    return [PitchClass(pc, 24) for pc in args]


def multiply(pcseg: list, n: int) -> list:
    """
    Multiplies a pcseg.
    :param pcseg: The pcseg
    :param n: The multiplier
    :return: The multiplied pcseg
    *Compatible with all PitchClass modulos    
    """
    pcseg2 = []
    if len(pcseg) > 0:
        mod = pcseg[0].mod
        for pc in pcseg:
            pcseg2.append(PitchClass(pc.pc * n, mod))
    return pcseg2


def multiply_order(pcseg: list, n: int) -> list:
    """
    Multiplies a pcseg's order.
    :param pcseg: The pcseg
    :param n: The multiplier
    :return: The order-multiplied pcseg
    *Compatible with all PitchClass modulos
    """
    pcseg2 = []
    if len(pcseg) > 0 and n > 0:
        mod = pcseg[0].mod
        for i in range(len(pcseg)):
            pcseg2.append(PitchClass(pcseg[(i * n) % len(pcseg)].pc, mod))
    return pcseg2


def ormap(row: list, ormap: dict) -> list:
    """
    Performs the ORMAP mapping on a row, given a provided ORMAP.
    :param row: A row
    :param ormap:
    :return: The ORMAP mapping for the row
    *Compatible with all PitchClass modulos
    """
    mapping = []
    for item in row:
        mapping.append(ormap[item])
    return mapping


def prot(pcseg: list) -> set:
    """
    Generates the protocol pairs for a pcseg.
    :param pcseg: A pcseg
    :return: A set of protocol pairs
    *Compatible with all PitchClass modulos
    """
    pp = set()
    for i in range(len(pcseg)):
        for j in range(i, len(pcseg)):
            pp.add((pcseg[i], pcseg[j]))
    return pp


def retrograde(pcseg: list) -> list:
    """
    Retrogrades a pcseg.
    :param pcseg: The pcseg
    :return: The retrograded pcseg
    *Compatible with all PitchClass modulos
    """
    pcseg2 = []
    if len(pcseg) > 0:
        mod = pcseg[0].mod
        for i in range(len(pcseg) - 1, -1, -1):
            pcseg2.append(PitchClass(pcseg[i].pc, mod))
    return pcseg2


def rotate(pcseg: list, n: int) -> list:
    """
    Rotates a pcseg.
    :param pcseg: The pcseg
    :param n: The index of rotation
    :return: The rotated pcseg
    *Compatible with all PitchClass modulos
    """
    pcseg2 = []
    if len(pcseg) > 0:
        mod = pcseg[0].mod
        for i in range(len(pcseg)):
            pcseg2.append(PitchClass(pcseg[(i - n) % len(pcseg)].pc, mod))
    return pcseg2


def transform(pcseg: list, string: str) -> list:
    """
    Transforms a pcseg with a provided transformation string.
    - Tn: transpose
    - I: invert
    - Mn: multiply
    - R: retrograde
    - rn: rotate
    :param pcseg: The pcseg to transform
    :param string: The transformation string
    :return: The transformed pcseg
    """
    pcseg2 = list.copy(pcseg)
    i = len(string) - 1
    while i >= 0:
        num = 0
        place_exp = 0
        while str.isdigit(string[i]) and i > 0:
            num += int(string[i]) * 10 ** place_exp
            place_exp += 1
            i -= 1
        match string[i]:
            case 'T':
                pcseg2 = transpose(pcseg2, num)
            case 'I':
                pcseg2 = invert(pcseg2)
            case 'M':
                pcseg2 = multiply(pcseg2, num)
            case 'R':
                pcseg2.reverse()
            case 'r':
                pcseg2 = rotate(pcseg2, num)
        i -= 1
    return pcseg2


def transform_hstack(pcseg: list, transformation_list: list) -> list:
    """
    Generates a really long pcseg by horizontally stacking transformed versions of it.
    :param pcseg: The pcseg to stack
    :param transformation_list: A list of transformations to stack
    :return: The stacked and transformed pcseg
    """
    pcseg1 = []
    for t in transformation_list:
        pcseg1 += transform(pcseg, t)
    return pcseg1


def transpose(pcseg: list, n: int) -> list:
    """
    Transposes a pcseg.
    :param pcseg: The pcseg
    :param n: The index of transposition
    :return: The transposed pcseg
    *Compatible with all PitchClass modulos
    """
    pcseg2 = []
    if len(pcseg) > 0:
        mod = pcseg[0].mod
        for pc in pcseg:
            pcseg2.append(PitchClass(pc.pc + n, mod))
    return pcseg2


def transpositional_combination(pcseg1: list, pcseg2: list) -> list:
    """
    Transpositionally combines (TC) two pcsegs. This is Boulez's "multiplication."
    :param pcseg1: A pcseg
    :param pcseg2: A pcseg
    :return: The TC pcset
    *Compatible with all PitchClass modulos
    """
    pcseg3 = []
    if len(pcseg1) > 0 and len(pcseg2) > 0:
        mod = pcseg1[0].mod
        for pc2 in pcseg2:
            for pc1 in pcseg1:
                pcseg3.append(PitchClass(pc1.pc + pc2.pc, mod))
    return pcseg3


class InvarianceMatrix:
    """
    Represents an invariance matrix. Compatible with all PitchClass modulos for T and I
    matrices, and compatible with mod 12 and mod 24 for various M matrices
    (M5, M7 for mod 12; M5, M7, M11, M13, M17, M19 for mod 24)
    """
    def __init__(self, a: list, c: list, mx_type="T"):
        """
        Creates an invariance matrix.
        :param a: Pcseg A
        :param c: Pcseg B
        :param mx_type: The matrix type (T, I, M, or MI)
        """
        self._a = None
        self._b = None
        self._c = None
        self._mx_type = mx_type.upper()
        self._mx = None
        self._load_matrix(a, c)

    def __getitem__(self, i: int, j: int) -> PitchClass:
        """
        Gets the pc at the specified row and column.
        :param i: The row
        :param j: The column
        :return: The pc
        """
        return self._mx[i][j]

    def __repr__(self):
        """
        Gets a representation of the InvarianceMatrix object.
        :returns: A string representation of the InvarianceMatrix object
        """
        return f"<pctheory.pcseg.InvarianceMatrix object at {id(self)}>: {self._mx}"

    def __str__(self):
        """
        Converts the InvarianceMatrix object to string.
        :returns: A string conversion of the InvarianceMatrix object
        """
        chars = ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B']
        lines = "    "
        for pc in self._a:
            lines += chars[pc.pc] + " "
        lines += "\n   "
        for i in range(2 * len(self._a)):
            lines += "-"
        lines += "\n"
        for i in range(len(self._mx) - 1):
            lines += chars[self._b[i].pc] + " | "
            for j in range(len(self._mx[i])):
                lines += chars[self._mx[i][j].pc] + " "
            lines += "\n"
        lines += chars[self._b[len(self._mx) - 1].pc] + " | "
        for j in range(len(self._mx[len(self._mx) - 1])):
            lines += chars[self._mx[len(self._mx) - 1][j].pc] + " "
        return lines

    @property
    def mx(self) -> list:
        """
        Gets the invariance matrix.
        :return: The invariance matrix
        """
        return self._mx

    @property
    def pcseg_a(self) -> list:
        """
        Gets the pcseg A.
        :return: The pcseg A
        """
        return self._a

    @property
    def pcseg_b(self) -> list:
        """
        Gets the pcseg B.
        :return: The pcseg B
        """
        return self._b

    @property
    def pcseg_c(self) -> list:
        """
        Gets the pcseg C.
        :return: The pcseg C
        """
        return self._c

    def get_column(self, j: int) -> list:
        """
        Gets a column of the matrix.
        :param j: The column index
        :return: The column
        """
        c = []
        for i in range(len(self._mx)):
            c.append(self._mx[i][j])
        return c

    def get_row(self, i: int) -> list:
        """
        Gets a row of the matrix.
        :param i: The row index
        :return: The row
        """
        return self._mx[i]

    def _load_matrix(self, a: list, c: list):
        """
        Loads the matrix.
        :param a: Pcseg A
        :param c: Pcseg C
        :return: None
        """
        if len(a) > 0 and len(c) > 0:
            ro = transformations.OTO()
            if a[0].mod == 12:
                INVERT = 11
                if self._mx_type == "T":
                    ro.oto = [0, 0, (1 * INVERT) % 12]
                elif self._mx_type == "M" or self._mx_type == "M5":
                    ro.oto = [0, 0, (5 * INVERT) % 12]
                elif self._mx_type == "MI" or self._mx_type == "M7":
                    ro.oto = [0, 0, (7 * INVERT) % 12]
                elif self._mx_type == "I" or self._mx_type == "M11":
                    ro.oto = [0, 0, (11 * INVERT) % 12]
                b = ro.transform(c)
            elif a[0].mod == 24:
                INVERT = 23
                if self._mx_type == "T":
                    ro.oto = [0, 0, (1 * INVERT) % 24]
                elif self._mx_type == "M5":
                    ro.oto = [0, 0, (5 * INVERT) % 24]
                elif self._mx_type == "M7":
                    ro.oto = [0, 0, (7 * INVERT) % 24]
                elif self._mx_type == "M11":
                    ro.oto = [0, 0, (11 * INVERT) % 24]
                elif self._mx_type == "M13":
                    ro.oto = [0, 0, (13 * INVERT) % 24]
                elif self._mx_type == "M17":
                    ro.oto = [0, 0, (17 * INVERT) % 24]
                elif self._mx_type == "M19":
                    ro.oto = [0, 0, (19 * INVERT) % 24]
                elif self._mx_type == "I" or self._mx_type == "M23":
                    ro.oto = [0, 0, (23 * INVERT) % 24]
                b = ro.transform(c)
            else:
                if self._mx_type == "T":
                    ro.oto = [0, 0, 1 * (a[0].mod - 1) % a[0].mod]
                elif self._mx_type == "I":
                    ro.oto = [0, 0, 11 * (a[0].mod - 1) % a[0].mod]
                b = ro.transform(c)

            self._mx = []
            for i in range(len(b)):
                mxrow = []
                for j in range(len(a)):
                    mxrow.append(PitchClass(b[i].pc + a[j].pc, a[0].mod))
                self._mx.append(mxrow)
            self._a = a.copy()
            self._b = b
            self._c = c.copy()

    def print(self, include: list = None):
        """
        Prints the invariance matrix.
        :param include: The pcs to include (if None, all pcs will be printed)
        :returns: None
        """
        chars = ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B']
        lines = "    "
        for pc in self._a:
            lines += chars[pc.pc] + " "
        lines += "\n   "
        for i in range(2 * len(self._a)):
            lines += "-"
        lines += "\n"
        if include is None:
            for i in range(len(self._mx) - 1):
                lines += chars[self._b[i].pc] + " | "
                for j in range(len(self._mx[i])):
                    lines += chars[self._mx[i][j].pc] + " "
                lines += "\n"
            lines += chars[self._b[len(self._mx) - 1].pc] + " | "
            for j in range(len(self._mx[len(self._mx) - 1])):
                lines += chars[self._mx[len(self._mx) - 1][j].pc] + " "
        else:
            for i in range(len(self._mx) - 1):
                lines += chars[self._b[i].pc] + " | "
                for j in range(len(self._mx[i])):
                    if self._mx[i][j] in include:
                        lines += chars[self._mx[i][j].pc] + " "
                    else:
                        lines += "  "
                lines += "\n"
            lines += chars[self._b[len(self._mx) - 1].pc] + " | "
            for j in range(len(self._mx[len(self._mx) - 1])):
                if self._mx[len(self._mx) - 1][j] in include:
                    lines += chars[self._mx[len(self._mx) - 1][j].pc] + " "
                else:
                    lines += "  "
        print(lines)


class TwelveToneMatrix:
    """
    Represents a twelve-tone matrix. Compatible only with mod 12 PitchClasses.
    """
    def __init__(self, row=None):
        """
        Creates a twelve-tone matrix.
        :param row: A row to import
        """
        self._labels_left = []
        self._labels_right = []
        self._labels_top = []
        self._labels_bottom = []
        self._mx = []
        self._row = None
        if row is not None:
            self.import_row(row)

    def __getitem__(self, i: int, j: int) -> PitchClass:
        """
        Gets the pc at the specified row and column.
        :param i: The row
        :param j: The column
        :return: The pc
        """
        return self._mx[i][j]

    def __repr__(self):
        """
        Generates a string representation of the TwelveToneMatrix that can be printed.
        :return: A string representation of the TwelveToneMatrix
        """
        lines = "     "
        for i in range(12):
            lines += f"{self._labels_top[i]: <5} "
        lines += "\n"
        for i in range(12):
            lines += f"{self._labels_left[i]: <4}"
            for j in range(12):
                lines += f"  {str(self._mx[i][j])}   "
            lines += f"{self._labels_right[i]: >4}"
            lines += "\n"
        lines += "     "
        for i in range(12):
            lines += f"{self._labels_bottom[i]: <5} "
        return lines

    def __str__(self):
        """
        Generates a string representation of the TwelveToneMatrix that can be printed.
        :return: A string representation of the TwelveToneMatrix
        """
        lines = "     "
        for i in range(12):
            lines += f"{self._labels_top[i]: <5} "
        lines += "\n"
        for i in range(12):
            lines += f"{self._labels_left[i]: <4}"
            for j in range(12):
                lines += f"  {str(self._mx[i][j])}   "
            lines += f"{self._labels_right[i]: >4}"
            lines += "\n"
        lines += "     "
        for i in range(12):
            lines += f"{self._labels_bottom[i]: <5} "
        return lines

    @property
    def labels_bottom(self) -> list:
        """
        Gets the bottom matrix labels.
        :return: The bottom matrix labels
        """
        return self._labels_bottom

    @property
    def labels_left(self) -> list:
        """
        Gets the left matrix labels.
        :return: The left matrix labels
        """
        return self._labels_left

    @property
    def labels_right(self) -> list:
        """
        Gets the right matrix labels.
        :return: The right matrix labels
        """
        return self._labels_right

    @property
    def labels_top(self) -> list:
        """
        Gets the top matrix labels.
        :return: The top matrix labels
        """
        return self._labels_top

    @property
    def matrix(self) -> list:
        """
        Gets the matrix.
        :return: The matrix
        """
        return self._mx

    @property
    def row(self) -> list:
        """
        Gets the row.
        :return: The row
        """
        return self._row

    def get_column(self, j: int) -> list:
        """
        Gets a column of the matrix.
        :param j: The column index
        :return: The column
        """
        pcseg = []
        for i in range(len(self._mx)):
            pcseg.append(PitchClass(self._mx[i][j].pc, 12))
        return pcseg

    def get_row(self, i: int) -> list:
        """
        Gets a row of the matrix.
        :param i: The row index
        :return: The row
        """
        return self._mx[i].copy()

    def import_row(self, row: list):
        """
        Imports a row.
        :param row: The row to import
        :return:
        """
        # We need the starting pitch to be 0 for the P0 form
        self._row = transpose(row, -row[0].pc)
        inv = invert(self._row)

        for i in range(len(self._row)):
            self._mx.append(transpose(self._row, inv[i].pc))

        for i in range(len(self._row)):
            p_lbl = self._mx[i][0] - self._mx[0][0]
            i_lbl = self._mx[0][i] - self._mx[0][0]
            r_lbl = self._mx[i][len(self._row) - 1] - self._mx[0][len(self._row) - 1]
            ri_lbl = self._mx[len(self._row) - 1][i] - self._mx[len(self._row) - 1][0]
            self._labels_bottom.append(f"T{ri_lbl.pc}RI")
            self._labels_left.append(f"T{p_lbl.pc}")
            self._labels_right.append(f"T{r_lbl.pc}R")
            self._labels_top.append(f"T{i_lbl.pc}I")



"""
File: pseg.py
Author: Jeff Martin
Date: 11/1/2021

Copyright © 2021 by Jeffrey Martin. All rights reserved.
Email: jmartin@jeffreymartincomposer.com
Website: https://jeffreymartincomposer.com

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from .pitch import Pitch, PitchClass


def intervals(pseg: list) -> list:
    """
    Gets the ordered interval content of a pseg
    :param pseg: The pseg
    :return: The ordered interval content as a list
    *Compatible with all Pitch modulos
    """
    intlist = []
    for i in range(1, len(pseg)):
        intlist.append(pseg[i].p - pseg[i - 1].p)
    return intlist
    

def invert(pseg: list) -> list:
    """
    Inverts a pseg
    :param pseg: The pseg
    :return: The inverted pseg
    *Compatible with all Pitch modulos
    """
    pseg2 = []
    if len(pseg) > 0:
        mod = pseg[0].mod
        for p in pseg:
            pseg2.append(Pitch(p.p * -1, mod))
    return pseg2


def make_pseg12(*args) -> list:
    """
    Makes a pseg
    :param args: Ps
    :return: A pseg
    *Compatible only with chromatic psegs
    """
    if len(args) > 0:
        if type(args[0]) == list:
            args = args[0]
        return [Pitch(p, 12) for p in args]
    else:
        return []


def make_pseg24(*args) -> list:
    """
    Makes a pseg
    :param args: Ps
    :return: A pseg
    *Compatible only with microtonal psegs
    """
    if len(args) > 0:
        if type(args[0]) == list:
            args = args[0]
        return [Pitch(p, 24) for p in args]
    else:
        return []


def multiply_order(pseg: list, n: int) -> list:
    """
    Multiplies a pseg's order
    :param pseg: The pseg
    :param n: The multiplier
    :return: The order-multiplied pseg
    *Compatible with all Pitch modulos
    """
    pseg2 = []
    if len(pseg) > 0:
        mod = pseg[0].mod
        for i in range(len(pseg)):
            pseg2.append(Pitch(pseg[(i * n) % len(pseg)].p, mod))
    return pseg2


def retrograde(pseg: list) -> list:
    """
    Retrogrades a pseg
    :param pseg: The pseg
    :return: The retrograded pseg
    *Compatible with all Pitch modulos
    """
    pseg2 = []
    if len(pseg) > 0:
        mod = pseg[0].mod
    for i in range(len(pseg) - 1, -1, -1):
        pseg2.append(Pitch(pseg[i].pc, mod))
    return pseg2


def rotate(pseg: list, n: int) -> list:
    """
    Rotates a pseg
    :param pseg: The pseg
    :param n: The index of rotation
    :return: The rotated pseg
    *Compatible with all Pitch modulos
    """
    pseg2 = []
    if len(pseg) > 0:
        mod = pseg[0].mod
    for i in range(len(pseg)):
        pseg2.append(Pitch(pseg[(i - n) % len(pseg)].p, mod))
    return pseg2


def to_pcseg(pseg: list) -> list:
    """
    Makes a pcseg out of a pseg
    :param pseg: A pseg
    :return: A pcseg
    *Compatible with all Pitch modulos
    """
    if len(pseg) > 0:
        mod = pseg[0].mod
        return [PitchClass(p.pc, mod) for p in pseg]
    else:
        return []


def transpose(pseg: list, n: int) -> list:
    """
    Transposes a pseg
    :param pseg: The pseg
    :param n: The index of transposition
    :return: The transposed pseg
    *Compatible with all Pitch modulos
    """
    pseg2 = []
    if len(pseg) > 0:
        mod = pseg[0].mod
        for p in pseg:
            pseg2.append(Pitch(p.p + n, mod))
    return pseg2



"""
File: pcset.py
Author: Jeff Martin
Date: 10/30/2021

Copyright © 2021 by Jeffrey Martin. All rights reserved.
Email: jmartin@jeffreymartincomposer.com
Website: https://jeffreymartincomposer.com

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from networkx import DiGraph
import pyvis
from . import tables, transformations
from .pitch import PitchClass
import numpy as np
import re


name_tables = tables.create_tables_sc12()


class SetClass:
    """
    Represents a pc-set-class.
    """
    def __init__(self, pcset=None, pc_mod: int = 12):
        """
        Creates a SetClass.
        :param pcset: A pcset to initialize the SetClass
        """
        if pcset is not None and len(pcset) > 0:
            self._NUM_PC = next(iter(pcset)).mod
        else:
            self._NUM_PC = pc_mod
        self._dsym = self._NUM_PC * 2
        self._ic_vector = [0 for i in range(self._NUM_PC // 2)]
        self._ic_vector_long = [0 for i in range(self._NUM_PC // 2 + 1)]
        self._name_carter = None
        self._name_forte = None
        self._name_morris = "()[]"
        self._name_prime = "[]"
        self._num_forte = None
        self._pcset = set()
        self._weight_right = True
        if pcset is not None:
            self.pcset = pcset

    def __eq__(self, other):
        if type(other) == SetClass:
            return self.pcset == other.pcset and self._NUM_PC == other._NUM_PC
        else:
            raise TypeError("SetClasses can only be compared to other SetClasses.")

    def __hash__(self):
        return hash(self._name_prime)

    def __len__(self):
        return len(self._pcset)

    def __lt__(self, other):
        if type(other) == SetClass:
            if self._NUM_PC != other._NUM_PC:
                raise ArithmeticError("Cannot compare two SetClasses with different mod values for <.")
            elif self._NUM_PC == 12:
                return len(self) < len(other) or (len(self) == len(other) and self._num_forte < other._num_forte)
            else:
                return len(self._pcset) < len(other._pcset) and self._name_prime < other._name_prime
        else:
            raise TypeError("SetClasses can only be compared to other SetClasses.")

    def __ne__(self, other):
        if type(other) == SetClass:
            return self._pcset != other._pcset or self._NUM_PC != other._NUM_PC
        else:
            raise TypeError("SetClasses can only be compared to other SetClasses.")

    def __repr__(self):
        # return f"<pctheory.pcset.SetClass object at {(id(self))}>: {repr(self._pcset)}"
        if self._NUM_PC == 12:
            return self._name_morris
        elif self._NUM_PC == None or self._NUM_PC <= 0:
            return "()[]"
        else:
            return self._name_prime

    def __str__(self):
        if self._NUM_PC == 12:
            return self._name_morris
        elif self._NUM_PC == None or self._NUM_PC <= 0:
            return "()[]"
        else:
            return self._name_prime
    
    @property
    def derived_core(self) -> list:
        """
        Gets derived core associations.
        :return: The derived core associations (or None if not derived core)
        """
        global name_tables
        if self.name_prime in name_tables["carterDerivedCoreTable"]:
            return [name for name in name_tables["carterDerivedCoreTable"][self.name_prime]]
        else:
            return None

    @property
    def dsym(self) -> int:
        """
        Gets the degree of symmetry of the set-class.
        :return: The degree of symmetry
        """
        return self._dsym

    @property
    def ic_vector(self) -> list:
        """
        Gets the IC vector.
        :return: The IC vector
        """
        return self._ic_vector

    @property
    def ic_vector_long(self) -> list:
        """
        Gets the IC vector in long format.
        :return: The IC vector in long format
        """
        return self._ic_vector_long

    @property
    def ic_vector_str(self) -> str:
        """
        Gets the IC vector as a string.
        :return: The IC vector
        """
        global name_tables
        if self._NUM_PC < 16:
            s = "["
            for a in self._ic_vector:
                s += name_tables["hexChars"][a]
            s += "]"
            return s
        else:
            return str(self._ic_vector)

    @property
    def ic_vector_long_str(self) -> str:
        """
        Gets the IC vector in long format as a string.
        :return: The IC vector in long format
        """
        global name_tables
        if self._NUM_PC < 16:
            s = "["
            for a in self._ic_vector_long:
                s += name_tables["hexChars"][a]
            s += "]"
            return s
        else:
            return str(self._ic_vector)

    @property
    def is_z_relation(self) -> bool:
        """
        Whether or not this set-class is Z-related to another set-class.
        :return: A boolean
        """
        if self._name_forte is not None and "Z" in self._name_forte:
            return True
        return False
    
    @property
    def mod(self) -> int:
        """
        Gets the modulo for PCs in this SetClass.
        :return: The modulo
        """
        return self._NUM_PC

    @property
    def name_carter(self) -> str:
        """
        Gets the Carter name for a set-class.
        :return: The Carter name
        """
        return self._name_carter

    @property
    def name_forte(self) -> str:
        """
        Gets the Forte name for a set-class.
        :return: The Forte name
        """
        return self._name_forte

    @property
    def name_morris(self) -> str:
        """
        Gets the Morris name for a set-class.
        :return: The Morris name
        """
        return self._name_morris

    @property
    def name_prime(self) -> str:
        """
        Gets the prime-form name (Rahn) for a set-class.
        :return: The prime-form name
        """
        return self._name_prime

    @property
    def num_forte(self) -> int:
        """
        Get the number part of the Forte name.
        :return: The number part of the Forte name
        """
        return self._num_forte

    @property
    def pcset(self) -> set:
        """
        Gets the pcset prime form.
        :return: The pcset prime form
        """
        return self._pcset

    @pcset.setter
    def pcset(self, value):
        """
        Updates the pcset prime form based on an existing pcset or pcseg.
        :param value: The new pcset or pcseg
        :return:
        """
        if type(value) == set or type(value) == list:
            for item in value:
                if type(item) != PitchClass:
                    raise TypeError("Cannot import sets into a SetClass if they are not composed exclusively of PitchClass objects.")
                elif item.mod != self.mod:
                    raise ArithmeticError(f"Cannot import sets into a SetClass with a different PitchClass modulo. This SetClass has modulo {self.mod}. You tried to import with a modulo of {item.mod}.")
            self._pcset = SetClass.calculate_prime_form(value, self._weight_right, self._NUM_PC)
            self._make_names()
        else:
            raise TypeError("Cannot import types other than sets and lists into a SetClass.")

    @property
    def weight_right(self) -> bool:
        """
        Whether or not to weight from the right.
        :return: A Boolean
        """
        return self._weight_right

    @weight_right.setter
    def weight_right(self, value: bool):
        """
        Whether or not to weight from the right.
        :param value: A Boolean
        :return:
        """
        self._weight_right = value
        self._pcset = SetClass.calculate_prime_form(self._pcset, self._weight_right, self._NUM_PC)

    @staticmethod
    def calculate_prime_form(pcset: set, weight_from_right: bool = True, pc_mod: int=12) -> set:
        """
        Calculates the prime form of a pcset.
        :param pcset: The pcset
        :param weight_from_right: Whether or not to pack from the right
        :param pc_mod: The PitchClass mod value to use 
        :return: The prime form
        """
        prime_set = set()
        if len(pcset) > 0:
            lists_to_weight = []
            pclist = [pc.pc for pc in pcset]
            inverted = [pc * -1 % pc_mod for pc in pclist]
            prime_list = None

            # Add regular forms
            for i in range(len(pclist)):
                lists_to_weight.append([])
                for i2 in range(i, len(pclist)):
                    lists_to_weight[i].append(pclist[i2])
                for i2 in range(0, i):
                    lists_to_weight[i].append(pclist[i2])
                initial_pitch = lists_to_weight[i][0]
                for i2 in range(0, len(lists_to_weight[i])):
                    lists_to_weight[i][i2] -= initial_pitch
                    if lists_to_weight[i][i2] < 0:
                        lists_to_weight[i][i2] += pc_mod
                lists_to_weight[i].sort()

            # Add inverted forms
            for i in range(len(pclist)):
                lists_to_weight.append([])
                for i2 in range(i, len(pclist)):
                    lists_to_weight[i + len(pclist)].append(inverted[i2])
                for i2 in range(0, i):
                    lists_to_weight[i + len(pclist)].append(inverted[i2])
                initial_pitch = lists_to_weight[i + len(pclist)][0]
                for i2 in range(0, len(lists_to_weight[i])):
                    lists_to_weight[i + len(pclist)][i2] -= initial_pitch
                    if lists_to_weight[i + len(pclist)][i2] < 0:
                        lists_to_weight[i + len(pclist)][i2] += pc_mod
                lists_to_weight[i + len(pclist)].sort()

            # Weight lists
            if weight_from_right:
                prime_list = SetClass._weight_from_right(lists_to_weight, pc_mod)
            else:
                prime_list = SetClass._weight_left(lists_to_weight, pc_mod)

            # Create pcset
            for pc in prime_list:
                prime_set.add(PitchClass(pc, pc_mod))

        return prime_set

    def contains_abstract_subset(self, sc) -> bool:
        """
        Determines if a set-class is an abstract subset of this set-class.
        :param sc: A set-class
        :return: A boolean
        """
        if type(sc) != SetClass:
            raise TypeError(f"Cannot subset items of type {type(sc)} from items of type SetClass.")
        elif sc.mod != self.mod:
            raise ArithmeticError(f"Cannot subset SetClasses of modulo {sc.mod} from SetClasses of modulo {self.mod}.")
        else:
            t_sets = []
            tni_sets = invert(sc.pcset)
            for i in range(self._NUM_PC):
                t_sets.append(transpose(sc.pcset, i))
                t_sets.append(transpose(tni_sets, i))
            for pcs in t_sets:
                if pcs.issubset(self.pcset):
                    return True
            return False

    def get_abstract_complement(self) -> 'SetClass':
        """
        Gets the abstract complement of the SetClass.
        :return: The abstract complement SetClass
        """
        complement_set_class = SetClass(pc_mod=self._NUM_PC)
        complement_set_class.pcset = get_complement(self._pcset)
        return complement_set_class

    def get_invariance_vector(self) -> list:
        """
        Gets the invariance vector of the SetClass
        :return: The invariance vector, or None if the SetClass has a PitchClass modulo other than 12
        """
        if self._NUM_PC == 12:
            iv = [0, 0, 0, 0, 0, 0, 0, 0]
            c = get_complement(self._pcset)
            utos = transformations.get_utos12()
            for i in range(self._NUM_PC):
                h = [utos[f"T{i}"].transform(self._pcset), utos[f"T{i}I"].transform(self._pcset),
                    utos[f"T{i}M"].transform(self._pcset), utos[f"T{i}MI"].transform(self._pcset)]
                for j in range(4):
                    if h[j] == self._pcset:
                        iv[j] += 1
                    if h[j].issubset(c):
                        iv[4 + j] += 1
            return iv
        else:
            return None

    def get_abstract_subset_classes(self) -> set:
        """
        Gets a set of subset-classes contained in this SetClass.
        :return:
        """
        subset_pcsets = subsets(self._pcset)
        subset_classes = set()
        for s in subset_pcsets:
            subset_classes.add(SetClass(s, pc_mod=self._NUM_PC))
        return subset_classes

    def get_partition2_subset_classes(self) -> set:
        """
        Gets a set of set-class partitions of this SetClass.
        :return:
        """
        p2 = partitions2(self._pcset)
        partitions2_set = set()
        for partition in p2:
            partitions2_set.add((SetClass(partition[0]), SetClass(partition[1])))
        return partitions2_set

    @staticmethod
    def get_set_classes12(cardinalities: list=None) -> list:
        """
        Gets the chromatic set-classes.
        :param cardinalities: A list of cardinalities if you don't want the entire list of 224 set-classes
        :return: A list of the chromatic set-classes
        """
        set_classes = []
        for name in name_tables["forteToSetNameTable"]:
            if cardinalities is not None:
                split = name.split('-')
                if int(split[0]) in cardinalities:
                    sc = SetClass(pc_mod=12)
                    sc.load_from_name(name)
                    set_classes.append(sc)
            else:
                sc = SetClass(pc_mod=12)
                sc.load_from_name(name)
                set_classes.append(sc)
        return set_classes

    def get_z_relation(self) -> 'SetClass':
        """
        Gets the Z-relation of the SetClass.
        :return: The Z-relation of the SetClass
        """
        if self._NUM_PC != 12:
            raise ArithmeticError("Cannot get Z-related sets for SetClasses with PitchClass modulo other than 12.")
        else:
            global name_tables
            zset = SetClass()
            f = self.name_forte
            if "Z" in f:
                zset.load_from_name(name_tables["zNameTable"][f])
            return zset

    def is_all_combinatorial_hexachord(self) -> bool:
        """
        Whether or not the SetClass is an all-combinatorial hexachord.
        :return: True or False
        """
        if self._name_prime in name_tables["allCombinatorialHexachords"] and self._NUM_PC == 12:
            return True
        else:
            return False

    @staticmethod
    def is_valid_name(name: str) -> bool:
        """
        Determines if a chromatic (mod 12) set-class name is valid. Validates prime form, Forte, and Morris names.
        Prime form name format: [xxxx]
        Forte name format: x-x
        Morris name format: (x-x)[xxxx]
        :param name: The name
        :return: A boolean
        """
        global name_tables
        if "[" in name and "-" in name:
            name = name.split(")")
            name[0] = name[0].replace("(", "")
            if name[0] in name_tables["forteToSetNameTable"] and name[1] in name_tables["setToForteNameTable"]:
                return True
        elif "-" in name:
            if name in name_tables["forteToSetNameTable"]:
                return True
        elif name in name_tables["setToForteNameTable"] or name in name_tables["setToForteNameTableLeftPacking"]:
            return True
        return False

    def load_from_name(self, name: str):
        """
        Loads a set-class from a prime-form, Morris, or Forte name.
        :param name: The name
        """
        global name_tables
        morris_matcher = re.compile(r'\(?\d+-z?\d+\)?\[\d+\]', re.IGNORECASE)
        forte_matcher = re.compile(r'\d+-z?\d+', re.IGNORECASE)
        text_cleaner = re.compile(r'\(|\)|\[|\]|\s')
        low_mod_prime_form_matcher = re.compile(r'\[|\([0-9a-zA-Z]+\]|\)')
        high_mod_prime_form_matcher = re.compile(r'\[|\([0-9a-zA-Z]+\]|\)')
        name = name.upper()
        prime_form_name = name

        if self._NUM_PC == 12:
            prime_form_name = f"[{text_cleaner.sub('', name)}]"
            # If it's a Morris name
            if morris_matcher.search(name):
                prime_form_name = name.split("[")[1]

            # If it's a Forte name
            elif forte_matcher.search(name):
                # Allow Forte names with or without Z, since it isn't fair to expect people to memorize 
                # which set-classes have Z-relations
                name2 = name.split('-')
                name2 = "-Z".join(name2)
                if name in name_tables["forteToSetNameTable"]:
                    prime_form_name = name_tables["forteToSetNameTable"][name]
                elif name2 in name_tables["forteToSetNameTable"]:
                    prime_form_name = name_tables["forteToSetNameTable"][name2]
                else:
                    raise Exception("Invalid Forte name.")

            elif prime_form_name not in name_tables["setToForteNameTable"]:
                raise Exception("Invalid set-class name.")
            
            prime_form_chars = [c for c in text_cleaner.sub('', prime_form_name)]
            pcset = set([PitchClass(name_tables["hexToInt"][pn], self._NUM_PC) for pn in prime_form_chars])
            self.pcset = pcset
            if self.name_prime != f"[{text_cleaner.sub('', prime_form_name)}]":
                raise Exception("Invalid set-class name.")
        
        elif self._NUM_PC <= 16 and low_mod_prime_form_matcher.search(name):
            prime_form_name = text_cleaner.sub('', name)
            prime_form_chars = [c for c in prime_form_name]
            pcset = set([PitchClass(name_tables["hexToInt"][pn], self._NUM_PC) for pn in prime_form_chars])
            self.pcset = pcset
            if self.name_prime != f"[{text_cleaner.sub('', name)}]":
                raise Exception("Invalid set-class name.")

        elif self._NUM_PC > 16:
            name = text_cleaner(name)
            name = name.split(",")
            pcset = set([PitchClass(int(chunk), self._NUM_PC) for chunk in name])
            self.pcset = pcset
        
        else:
            raise Exception("Invalid set-class name.")

    def _make_names(self):
        """
        Makes the names for the set-class.
        :return:
        """
        global name_tables
        pc_list = [pc.pc for pc in self._pcset]
        pc_list.sort()
        self._name_carter = None
        self._name_morris = None
        self._name_forte = None
        self._num_forte = None

        if self._NUM_PC <= 16:
            self._name_prime = "[" + "".join([name_tables["hexChars"][pc] for pc in pc_list]) + "]"
        else:
            self._name_prime = "[" + ", ".join([f"{pc:0>2}" for pc in pc_list]) + "]"

        if self._NUM_PC == 12:
            if self._name_prime != "[]":
                if self._name_prime in name_tables["setToForteNameTableLeftPacking"]:
                    self._name_forte = name_tables["setToForteNameTableLeftPacking"][self._name_prime]
                else:
                    self._name_forte = name_tables["setToForteNameTable"][self._name_prime]
            else:
                self._name_forte = "0-1"
            self._name_carter = ""
            if self._name_forte in name_tables["forteToCarterNameTable"]:
                self._name_carter = name_tables["forteToCarterNameTable"][self._name_forte]
            self._name_morris = "(" + self._name_forte + ")" + self._name_prime
            forte_num = self.name_forte.split('-')[1]
            forte_num = forte_num.strip('Z')
            self._num_forte = int(forte_num)

        # Calculate the IC vector
        self._ic_vector_long = [0 for i in range(self._NUM_PC // 2 + 1)]
        for pc in self._pcset:
            for pc2 in self._pcset:
                interval = (pc2.pc - pc.pc) % self._NUM_PC
                if interval > self._NUM_PC // 2:
                    interval = interval * -1 + self._NUM_PC
                self._ic_vector_long[interval] += 1
        for i in range(1, len(self._ic_vector_long)):
            self._ic_vector_long[i] //= 2
        self._ic_vector = self._ic_vector_long[1:]

        # Get the degree of symmetry
        if len(self._pcset) > 0:
            c = get_corpus(self._pcset)
            self._dsym = (self._NUM_PC * 2) // len(c)
        else:
            self._dsym = self._NUM_PC * 2

    @staticmethod
    def _weight_from_right(pclists: list, pc_mod: int=12):
        """
        Weights pclists from the right.
        :param pclists: Pclists
        :param pc_mod: The PitchClass mod value to use 
        :return: The most weighted form
        """
        for i in range(len(pclists[0]) - 1, -1, -1):
            if len(pclists) > 1:
                # The smallest item at the current index
                smallest_item = pc_mod - 1

                # Identify the smallest item at the current index
                for j in range(len(pclists)):
                    if pclists[j][i] < smallest_item:
                        smallest_item = pclists[j][i]

                # Remove all lists with larger items at the current index
                j = 0
                while j < len(pclists):
                    if pclists[j][i] > smallest_item:
                        del pclists[j]
                    else:
                        j += 1

            else:
                break
        return pclists[0]

    @staticmethod
    def _weight_left(pclists: list, pc_mod: int=12):
        """
        Weights pclists left.
        :param pclists: Pclists
        :param pc_mod: The PitchClass mod value to use 
        :return: The most weighted form
        """
        if len(pclists) > 1:
            # The smallest item at the current index
            smallest_item = pc_mod - 1

            # Identify the smallest item at the last index
            for j in range(0, len(pclists)):
                if pclists[j][len(pclists[0]) - 1] < smallest_item:
                    smallest_item = pclists[j][len(pclists[0]) - 1]

            # Remove all lists with larger items at the current index
            j = 0
            while j < len(pclists):
                if pclists[j][len(pclists[0]) - 1] > smallest_item:
                    del pclists[j]
                else:
                    j += 1

            # Continue processing, but now pack from the left
            for i in range(0, len(pclists[0])):
                if len(pclists) > 1:
                    smallest_item = pc_mod - 1

                    # Identify the smallest item at the current index
                    for j in range(len(pclists)):
                        if pclists[j][i] < smallest_item:
                            smallest_item = pclists[j][i]

                    # Remove all lists with larger items at the current index
                    j = 0
                    while j < len(pclists):
                        if pclists[j][i] > smallest_item:
                            del pclists[j]
                        else:
                            j += 1
                else:
                    break
        return pclists[0]


def get_all_combinatorial_hexachord(name: str) -> SetClass:
    """
    Gets an all-combinatorial hexachord (ACH) by name (A-F).
    :param name: The name of the hexachord (A-F)
    :return: The hexachord set-class
    *Only produces mod 12 SetClasses
    """
    sc = SetClass(pc_mod=12)
    sc.load_from_name(name_tables["allCombinatorialHexachordNames"][name])
    return sc


def get_complement(pcset: set) -> set:
    """
    Gets the complement of a pcset.
    :param pcset: A pcset
    :return: The complement pcset
    *Compatible with all PitchClass modulos
    """
    universal = set()
    if len(pcset) > 0:
        mod = next(iter(pcset)).mod
        for i in range(mod):
            universal.add(PitchClass(i, mod))
    return universal - pcset


def get_complement_map_utos(pcset: set) -> set:
    """
    Gets all UTOs that map a pcset into its complement.
    :param pcset: A pcset
    :return: A set of UTOs
    *Compatible with PitchClasses mod 12 and 24
    """
    utos = set()
    mod = next(iter(pcset)).mod
    c = get_complement(pcset)
    if mod == 12:
        uto = transformations.get_utos12()
        for i in range(12):
            tx = uto[f"T{i}"].transform(pcset)
            m5x = uto[f"T{i}M5"].transform(pcset)
            m7x = uto[f"T{i}M7"].transform(pcset)
            m11x = uto[f"T{i}M11"].transform(pcset)
            if tx.issubset(c):
                utos.add(uto[f"T{i}"])
            if m5x.issubset(c):
                utos.add(uto[f"T{i}M5"])
            if m7x.issubset(c):
                utos.add(uto[f"T{i}M7"])
            if m11x.issubset(c):
                utos.add(uto[f"T{i}M11"])
    else:
        uto = transformations.get_utos24()
        for i in range(24):
            tx = uto[f"T{i}"].transform(pcset)
            m5x = uto[f"T{i}M5"].transform(pcset)
            m7x = uto[f"T{i}M7"].transform(pcset)
            m11x = uto[f"T{i}M11"].transform(pcset)
            m13x = uto[f"T{i}M13"].transform(pcset)
            m17x = uto[f"T{i}M17"].transform(pcset)
            m19x = uto[f"T{i}M19"].transform(pcset)
            m23x = uto[f"T{i}M23"].transform(pcset)
            if tx.issubset(c):
                utos.add(uto[f"T{i}"])
            if m5x.issubset(c):
                utos.add(uto[f"T{i}M5"])
            if m7x.issubset(c):
                utos.add(uto[f"T{i}M7"])
            if m11x.issubset(c):
                utos.add(uto[f"T{i}M11"])
            if m13x.issubset(c):
                utos.add(uto[f"T{i}M13"])
            if m17x.issubset(c):
                utos.add(uto[f"T{i}M17"])
            if m19x.issubset(c):
                utos.add(uto[f"T{i}M19"])
            if m23x.issubset(c):
                utos.add(uto[f"T{i}M23"])
    return utos


def get_corpus(pcset: set) -> set:
    """
    Gets all transformations of a provided pcset.
    :param pcset: A pcset
    :return: A set of all transformations of the pcset
    *Compatible with all PitchClass modulos
    """
    pcsets = set()
    if len(pcset) > 0:
        mod = next(iter(pcset)).mod
        for i in range(mod):
            pcsets.add(frozenset(transpose(pcset, i)))
            pcsets.add(frozenset(transpose(invert(pcset), i)))
    return pcsets


def get_self_map_utos(pcset: set) -> set:
    """
    Gets all UTOs that map a pcset into itself.
    :param pcset: A pcset
    :return: A set of UTOs
    *Compatible with PitchClasses mod 12 and 24
    """
    utos = set()
    t = type(next(iter(pcset)))
    if t == PitchClass:
        uto = transformations.get_utos12()
        for i in range(12):
            tx = uto[f"T{i}"].transform(pcset)
            m5x = uto[f"T{i}M5"].transform(pcset)
            m7x = uto[f"T{i}M7"].transform(pcset)
            m11x = uto[f"T{i}M11"].transform(pcset)
            if tx == pcset:
                utos.add(uto[f"T{i}"])
            if m5x == pcset:
                utos.add(uto[f"T{i}M5"])
            if m7x == pcset:
                utos.add(uto[f"T{i}M7"])
            if m11x == pcset:
                utos.add(uto[f"T{i}M11"])
    else:
        uto = transformations.get_utos24()
        for i in range(24):
            tx = uto[f"T{i}"].transform(pcset)
            m5x = uto[f"T{i}M5"].transform(pcset)
            m7x = uto[f"T{i}M7"].transform(pcset)
            m11x = uto[f"T{i}M11"].transform(pcset)
            m13x = uto[f"T{i}M13"].transform(pcset)
            m17x = uto[f"T{i}M17"].transform(pcset)
            m19x = uto[f"T{i}M19"].transform(pcset)
            m23x = uto[f"T{i}M23"].transform(pcset)
            if tx == pcset:
                utos.add(uto[f"T{i}"])
            if m5x == pcset:
                utos.add(uto[f"T{i}M5"])
            if m7x == pcset:
                utos.add(uto[f"T{i}M7"])
            if m11x == pcset:
                utos.add(uto[f"T{i}M11"])
            if m13x == pcset:
                utos.add(uto[f"T{i}M13"])
            if m17x == pcset:
                utos.add(uto[f"T{i}M17"])
            if m19x == pcset:
                utos.add(uto[f"T{i}M19"])
            if m23x == pcset:
                utos.add(uto[f"T{i}M23"])
    return utos


def convert_to_pcset12(pcset: set) -> set:
    """
    Converts a microtonal pcset (mod 24) to a chromatic pcset (mod 12). Microtonal pitch classes
    are rounded down to the nearest chromatic pitch class.
    :param args: A microtonal pcset (mod 24)
    :return: A chromatic pcset (mod 12)
    """
    return {PitchClass(pc.pc // 2, 12) for pc in pcset}


def convert_to_pcset24(pcset: set) -> set:
    """
    Converts a chromatic pcset (mod 12) to a microtonal pcset (mod 24).
    :param args: A chromatic pcset (mod 12)
    :return: A microtonal pcset (mod 24)
    """
    return {PitchClass(pc.pc * 2, 24) for pc in pcset}
    

def invert(pcset: set) -> set:
    """
    Inverts a pcset.
    :param pcset: The pcset
    :return: The inverted pcset
    *Compatible with all PitchClass modulos
    """
    pcset2 = set()
    if len(pcset) > 0:
        for pc in pcset:
            pcset2.add(PitchClass(pc.pc * -1, pc.mod))
    return pcset2


def is_all_combinatorial_hexachord(pcset: set) -> bool:
    """
    Whether or not a pcset is an all-combinatorial hexachord.
    :param pcset: A pcset
    :return: True or False
    *Only compatible with mod 12 SetClasses
    """
    sc = SetClass(pcset)
    if sc.name_prime in name_tables["allCombinatorialHexachords"]:
        return True
    else:
        return False


def make_pcset12(*args) -> set:
    """
    Makes a chromatic pcset (mod 12).
    :param args: Integers that represent pitch classes
    :return: A pcset
    """
    if type(args[0]) == list:
        args = args[0]
    return {PitchClass(pc, 12) for pc in args}


def make_pcset24(*args) -> set:
    """
    Makes a microtonal pcset (mod 24).
    :param args: Integers that represent pitch classes
    :return: A pcset
    """
    if type(args[0]) == list:
        args = args[0]
    return {PitchClass(pc, 24) for pc in args}


def make_subset_graph(set_class: SetClass, smallest_cardinality: int = 1, show_graph: bool = False, size: tuple = (800, 1100)) -> DiGraph:
    """
    Makes a subset graph.
    :param set_class: A set-class
    :param smallest_cardinality: The smallest cardinality to include in the graph
    :param show_graph: Whether or not to generate a visualization of the graph
    :param size: The size of the visualized graph
    :return: A graph
    """
    subset_graph = DiGraph()
    set_classes = list(set_class.get_abstract_subset_classes())
    for sc in set_classes:
        if len(sc.pcset) >= smallest_cardinality:
            subset_graph.add_node(sc.name_prime)
    for i in range(0, len(set_classes)):
        for j in range(0, i):
            if set_classes[i].contains_abstract_subset(set_classes[j]) and len(set_classes[j].pcset) >= smallest_cardinality:
                subset_graph.add_edge(set_classes[i].name_prime, set_classes[j].name_prime)
        for j in range(i + 1, len(set_classes)):
            if set_classes[i].contains_abstract_subset(set_classes[j]) and len(set_classes[j].pcset) >= smallest_cardinality:
                subset_graph.add_edge(set_classes[i].name_prime, set_classes[j].name_prime)
    if show_graph:
        net = pyvis.network.Network(f"{size[0]}px", f"{size[1]}px", directed=True, bgcolor="#eeeeee",
                                    font_color="#333333", heading="Subset Graph")
        net.toggle_hide_edges_on_drag(False)
        net.barnes_hut()
        net.from_nx(subset_graph, default_node_size=40)
        for node in net.nodes:
            node["title"] = node["id"]
            node["color"] = "#41b535"
        for edge in net.edges:
            edge["color"] = "#1c4219"
        net.show("subset_graph.html")
    return subset_graph


def multiply(pcset: set, n: int) -> set:
    """
    Multiplies a pcset.
    :param pcset: The pcset
    :param n: The multiplier
    :return: The multiplied pcset
    *Compatible with all PitchClass modulos
    """
    pcset2 = set()
    if len(pcset) > 0:
        for pc in pcset:
            pcset2.add(PitchClass(pc.pc * n, pc.mod))
    return pcset2


def partitions2(pcset: set) -> list:
    """
    Gets all partitions of a pcset (size 2 or 1).
    :param pcset: A pcset
    :return: A list of all partitions
    *Compatible with all PitchClass modulos
    """
    subs = subsets(pcset)
    partitions_dict = {}
    partitions_list = []
    len_pcset = (len(pcset) + 1) // 2 if len(pcset) % 2 else len(pcset) // 2
    for sub in subs:
        if len(sub) <= len_pcset:
            s = frozenset(sub)
            d = frozenset(pcset.difference(sub))
            if d not in partitions_dict:
                partitions_dict[s] = d
    for s in partitions_dict:
        partitions_list.append((set(s), set(partitions_dict[s])))
    return partitions_list


def permutations(pcset: set) -> list:
    """
    Generates all permutations of a pcset. Uses a swapping notion derived from the Bauer-Mengelberg/Ferentz algorithm
    for generating all-interval twelve-tone rows.
    Note: The number of permutations will be n! where n is the length of the pcset. The amount of pcsegs is therefore
    O(n!). You may not want to try generating all permutations of a twelve-note set.
    You have been warned.
    :param pcs: A pcset
    :return: A list of pcsegs
    *Compatible with all PitchClass modulos
    """
    current_permutation = [pc.pc for pc in pcset]
    current_permutation.sort()
    
    # Determine the type of pitch-class, and the number of possible pitch-classes
    mod = next(iter(pcset)).mod
    
    all_permutations = []  # This will hold the final list of permutations
    critical_index = 0     # The index of the critical digit

    # This array keeps track of all numbers we have found past the critical index.
    flags = np.zeros((mod), dtype=np.int8)
    
    while critical_index > -1:
        all_permutations.append([PitchClass(pc, mod) for pc in current_permutation])
        critical_index = -1
        next_higher_digit = 0
        
        # Examine the permutation to find the critical digit
        for i in range(len(current_permutation) - 1, 0, -1):
            flags[current_permutation[i]] = 1
            
            # If we've found the critical index
            if current_permutation[i-1] < current_permutation[i]:
                flags[current_permutation[i-1]] = 1
                critical_index = i - 1
                
                # Find the next highest number so we can do the swap
                for j in range(current_permutation[critical_index] + 1, mod):
                    if flags[j]:
                        next_higher_digit = j
                        flags[j] = 0
                        break

                # Swap the critical digit
                current_permutation[critical_index] = next_higher_digit

                # Repopulate the permutation
                for j in range(mod):
                    if flags[j]:
                        critical_index += 1
                        current_permutation[critical_index] = j
                        flags[j] = 0
                break
    return all_permutations


def set_class_filter12(name: str, sets: list) -> list:
    """
    Filters a list of pcsets.
    :param name: The name to find
    :param sets: A list of sets to filter
    :return: A filtered list
    *Compatible with all PitchClass modulos. For pcsets of modulo 12, also supports Forte and Morris names.
    """
    newlist = []
    sc = SetClass()
    for s in sets:
        sc.pcset = s
        if sc.name_prime == name or sc.name_forte == name or sc.name_morris == name:
            newlist.append(s)
    return newlist


def subsets(pcset: set) -> list:
    """
    Gets all subsets of a pcset, using the bit masking solution from
    https://afteracademy.com/blog/print-all-subsets-of-a-given-set
    :param pcset: A pcset
    :return: A list containing all subsets of the pcset
    """
    total = 2 ** len(pcset)
    sub = []
    if total > 1:
        mod = next(iter(pcset)).mod
        pcseg = list(pcset)
        pcseg.sort()
        for index in range(total):
            sub.append([])
            for i in range(len(pcset)):
                if index & (1 << i):
                    sub[index].append(PitchClass(pcseg[i].pc, mod))
        sub.sort()
    return sub


def transform(pcset, string) -> set:
    """
    Transforms a pcset with a provided transformation string.
    - Tn: transpose
    - I: invert
    - Mn: multiply
    :param pcset: The pcset to transform
    :param string: The transformation string
    :return: The transformed pcset
    """
    pcset2 = set.copy(pcset)
    i = len(string) - 1
    while i >= 0:
        num = 0
        place_exp = 0
        while str.isdigit(string[i]) and i > 0:
            num += int(string[i]) * 10 ** place_exp
            place_exp += 1
            i -= 1
        match string[i]:
            case 'T':
                pcset2 = transpose(pcset2, num)
            case 'I':
                pcset2 = invert(pcset2)
            case 'M':
                pcset2 = multiply(pcset2, num)
        i -= 1
    return pcset2


def transpose(pcset: set, n: int) -> set:
    """
    Transposes a pcset.
    :param pcset: The pcset
    :param n: The index of transposition
    :return: The transposed pcset
    *Compatible with all PitchClass modulos
    """
    pcset2 = set()
    if len(pcset) > 0:
        mod = next(iter(pcset)).mod
        for pc in pcset:
            pcset2.add(PitchClass(pc.pc + n, mod))
    return pcset2


def transpositional_combination(pcset1: set, pcset2: set) -> set:
    """
    Transpositionally combines (TC) two pcsets. This is Boulez's "multiplication."
    :param pcset1: A pcset
    :param pcset2: A pcset
    :return: The TC pcset
    *Compatible with all PitchClass modulos
    """
    pcset3 = set()
    if len(pcset1) > 0 and len(pcset2) > 0:
        mod = next(iter(pcset1)).mod
        for pc2 in pcset2:
            for pc1 in pcset1:
                pcset3.add(PitchClass(pc1.pc + pc2.pc, mod))
    return pcset3


def visualize(pcset: set) -> str:
    """
    Visualizes a pcset.
    :param pcset: A pcset
    :return: A visualization
    *Compatible with all PitchClass modulos
    """
    line = ""
    if len(pcset) > 0:
        mod = next(iter(pcset)).mod
        for i in range(mod):
            if PitchClass(i, mod) in pcset:
                line += "X"
            else:
                line += " "
    return line



"""
File: pitch.py
Author: Jeff Martin
Date: 10/30/2021

Copyright © 2021, 2023 by Jeffrey Martin. All rights reserved.
Email: jmartin@jeffreymartincomposer.com
Website: https://www.jeffreymartincomposer.com

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

_hex_map = {0: '0', 1: '1', 2: '2', 3: '3', 4: '4', 5: '5', 6: '6', 7: '7', 8: '8', 9: '9', 10: 'A', 11: 'B', 12: 'C', 13: 'D', 14: 'E', 15: 'F'}


class PitchClass:
    def __init__(self, pc: int=0, mod: int=12):
        """
        Creates a PitchClass
        :param pc: The pitch class integer
        :param mod: The mod number (12 for default chromatic pitch-classes, 
        24 for quarter-tone pitch-classes)
        """
        self._mod = mod
        self._pc = pc % mod

    def __add__(self, other):
        if type(other) == PitchClass:
            if self._mod == other._mod:
                return PitchClass(self.pc + other.pc, self._mod)
            else:
                raise ArithmeticError("Cannot add two pitch-classes with different mod values.")
        elif type(other) == int:
            return PitchClass(self.pc + other, self._mod)
        else:
            raise TypeError("PitchClasses can only be added to other PitchClasses of the same modulo, or to integers.")

    def __eq__(self, other):
        if type(other) == PitchClass:
            return self.pc == other.pc and self._mod == other._mod
        else:
            raise TypeError("PitchClasses can only be compared to other PitchClasses.")

    def __ge__(self, other):
        if type(other) == PitchClass:
            if self._mod == other._mod:
                return self.pc >= other.pc
            else:
                raise ArithmeticError("Cannot compare two pitch-classes with different mod values for >=.")
        else:
            raise TypeError("PitchClasses can only be compared to other PitchClasses.")

    def __gt__(self, other):
        if type(other) == PitchClass:
            if self._mod == other._mod:
                return self.pc > other.pc
            else:
                raise ArithmeticError("Cannot compare two pitch-classes with different mod values for >.")
        else:
            raise TypeError("PitchClasses can only be compared to other PitchClasses.")

    def __hash__(self):
        return hash(f"pc{self._pc}-{self._mod}")

    def __le__(self, other):
        if type(other) == PitchClass:
            if self._mod == other._mod:
                return self.pc <= other.pc
            else:
                raise ArithmeticError("Cannot compare two pitch-classes with different mod values for <=.")
        else:
            raise TypeError("PitchClasses can only be compared to other PitchClasses.")

    def __lt__(self, other):
        if type(other) == PitchClass:
            if self._mod == other._mod:
                return self.pc < other.pc
            else:
                raise ArithmeticError("Cannot compare two pitch-classes with different mod values for <.")
        else:
            raise TypeError("PitchClasses can only be compared to other PitchClasses.")

    def __mul__(self, other):
        if type(other) == PitchClass:
            if self._mod == other._mod:
                return PitchClass(self.pc * other.pc, self._mod)
            else:
                raise ArithmeticError("Cannot multiply two pitch-classes with different mod values.")
        elif type(other) == int:
            return PitchClass(self.pc * other, self._mod)
        else:
            raise TypeError("PitchClasses can only be multiplied by other PitchClasses of the same modulo, or by integers.")

    def __ne__(self, other):
        if type(other) == PitchClass:
            return self.pc != other.pc or self._mod != other._mod
        else:
            raise TypeError("PitchClasses can only be compared to other PitchClasses.")

    def __repr__(self):
        # return "<pctheory.pitch.PitchClass object at " + str(id(self)) + ">: " + self.pc_char
        return self.pc_str

    def __str__(self):
        return self.pc_str

    def __sub__(self, other):
        if type(other) == PitchClass:
            if self._mod == other._mod:
                return PitchClass(self.pc - other.pc, self._mod)
            else:
                raise ArithmeticError("Cannot subtract two pitch-classes with different mod values.")
        elif type(other) == int:
            return PitchClass(self.pc - other, self._mod)
        else:
            raise TypeError("PitchClasses can only be subtracted by other PitchClasses of the same modulo, or by integers.")

    @property
    def mod(self):
        """
        The pitch-class modulo
        :return: The pitch-class modulo
        """
        return self._mod

    @property
    def pc(self) -> int:
        """
        The pitch-class integer
        :return: The pitch-class integer
        """
        return self._pc

    @pc.setter
    def pc(self, value: int):
        """
        The pitch-class integer
        :param value: The new pitch-class integer
        :return:
        """
        self._pc = value % 12
        if self._pc < 0:
            self._pc += 12

    @property
    def pc_str(self):
        """
        The pitch-class string
        :return: The pitch-class string
        """
        if 10 < self._mod <= 16:
            return _hex_map[self._pc]
        elif self.mod > 16:
            return f"{self._pc:0>2}"
        else:
            return str(self._pc)
        

class Pitch(PitchClass):
    """
    Represents a pitch
    """
    def __init__(self, p: int=0, pc_mod: int=12, pname: str=0):
        """
        Creates a Pitch
        :param p: The pitch integer
        :param pc_mod: The modulo for the underlying PitchClass
        :param pname: The pitch name as a string (optional, for display purposes)
        """
        super().__init__(p, pc_mod)
        self._p = p
        self._pname = pname

    def __add__(self, other):
        if type(other) == Pitch:
            if self._mod == other._mod:
                return Pitch(self.p + other.p, self._mod)
            else:
                raise ArithmeticError("Cannot add two Pitches with different PitchClass mod values.")
        elif type(other) == int:
            return Pitch(self.p + other, self._mod)
        else:
            raise TypeError("Pitches can only be added to other Pitches of the same PitchClass modulo, or to integers.")

    def __eq__(self, other):
        if type(other) == Pitch:
            return self._p == other._p and self._mod == other._mod
        else:
            raise TypeError("Pitches can only be compared to other Pitches.")

    def __ge__(self, other):
        if type(other) == Pitch:
            if self._mod == other._mod:
                return self.p >= other.p
            else:
                raise ArithmeticError("Cannot compare two Pitches with different PitchClass mod values for >=.")
        else:
            raise TypeError("Pitches can only be compared to other Pitches.")

    def __gt__(self, other):
        if type(other) == Pitch:
            if self._mod == other._mod:
                return self.p > other.p
            else:
                raise ArithmeticError("Cannot compare two Pitches with different PitchClass mod values for >.")
        else:
            raise TypeError("Pitches can only be compared to other Pitches.")

    def __hash__(self):
        return hash(f"p{self._p}-{self._mod}")

    def __le__(self, other):
        if type(other) == Pitch:
            if self._mod == other._mod:
                return self.p <= other.p
            else:
                raise ArithmeticError("Cannot compare two Pitches with different PitchClass mod values for <=.")
        else:
            raise TypeError("Pitches can only be compared to other Pitches.")

    def __lt__(self, other):
        if type(other) == Pitch:
            if self._mod == other._mod:
                return self.p < other.p
            else:
                raise ArithmeticError("Cannot compare two Pitches with different PitchClass mod values for <.")
        else:
            raise TypeError("Pitches can only be compared to other Pitches.")

    def __mul__(self, other):
        if type(other) == Pitch:
            if self._mod == other._mod:
                return Pitch(self.p * other.p, self._mod)
            else:
                raise ArithmeticError("Cannot multiply two Pitches with different PitchClass mod values.")
        elif type(other) == int:
            return Pitch(self.p * other, self._mod)
        else:
            raise TypeError("Pitches can only be multiplied by other Pitches of the same PitchClass modulo, or by integers.")

    def __ne__(self, other):
        if type(other) == Pitch:
            return self._p != other._p or self._mod != other._mod
        else:
            raise TypeError("Pitches can only be compared to other Pitches.")

    def __repr__(self):
        # return "<pctheory.pitch.Pitch object at " + str(id(self)) + ">: " + str(self._p)
        return str(self._p)

    def __str__(self):
        return str(self._p)

    def __sub__(self, other):
        if type(other) == Pitch:
            if self._mod == other._mod:
                return Pitch(self.p - other.p, self._mod)
            else:
                raise ArithmeticError("Cannot subtract two Pitches with different PitchClass mod values.")
        elif type(other) == int:
            return Pitch(self.p - other, self._mod)
        else:
            raise TypeError("Pitches can only be subtracted by other Pitches of the same PitchClass modulo, or by integers.")

    @property
    def p(self) -> int:
        """
        The pitch integer
        :return: The pitch integer
        """
        return self._p

    @p.setter
    def p(self, value: int):
        """
        The pitch integer
        :param value: The new pitch integer
        :return:
        """
        self._p = value
        self.pc = self._p

    @property
    def pname(self) -> str:
        """
        The pitch name of the pitch
        :return: The pitch name
        """
        return self._pname

    @pname.setter
    def pname(self, value: str):
        """
        The pitch name of the pitch
        :param value: The new name
        :return:
        """
        self._pname = value

    @property
    def midi(self) -> int:
        """
        The MIDI version of the pitch integer
        :return: The MIDI version of the pitch integer. If the pitch is 
        microtonal or anything other than a 12-tone chromatic pitch, 
        the MIDI value will be an appropriate floating-point number.
        """
        if self._mod == 12:
            return 60 + self._p
        else:
            return 60 + self._p // self._mod * 12 + self._pc * 12 / self._mod

    @midi.setter
    def midi(self, value):
        """
        The MIDI version of the pitch integer
        :param value: The new MIDI version of the pitch integer
        :return:
        """
        self._p = value - 60
        self.pc = self._p
